import os
import pickle
import multiprocessing
import itertools
import shutil
import subprocess


import pandas as pnd
from Bio import SeqIO, SeqRecord, Seq


from ..commons import check_cached
from ..commons import chunkize_items
from ..commons import load_the_worker
from ..commons import gather_results
from ..commons import get_blast_header
from ..commons import extract_aa_seq_from_genome
from ..commons import create_summary
from ..commons import update_pam



def update_seq_to_coords(logger): 
    
    
    # load the previously created species_to_proteome: 
    with open('working/proteomes/species_to_proteome.pickle', 'rb') as handler:
        species_to_proteome = pickle.load(handler)
        
    
    # get the accessions (passing the quality filters):
    accessions = []
    for species in species_to_proteome.keys(): 
        for proteome in species_to_proteome[species]:
            basename = os.path.basename(proteome)
            accession, _ = os.path.splitext(basename)
            accessions.append(accession)
    

    # create an updateed seq_to_coords dict: 
    with open('working/rec_masking/seq_to_coords.pickle', 'rb') as handler:
        seq_to_coords_update = pickle.load(handler)
    logger.debug(f'rec_overlap: seq_to_coords: starting from {len(seq_to_coords_update.values())} sequences.')

        
    # now add the new seqs (recovered by this module): 
    for accession in accessions: 
        results_df = pnd.read_csv(f'working/rec_overlap/results/{accession}.csv', index_col=0)
        for index, row in results_df.iterrows():
            seq_to_coords_update[row['ID']] = {'accession': row['accession'], 'contig': row['contig'], 'strand': row['strand'], 'start': row['start'], 'end': row['end']}
    logger.debug(f'rec_overlap: seq_to_coords: {len(seq_to_coords_update.values())} sequences after the addition of new IDs.')

    
    # save the update dictionary: 
    with open('working/rec_overlap/seq_to_coords.pickle', 'wb') as file:
        pickle.dump(seq_to_coords_update, file)



def task_recoverlap(genome, args):
    
    
    # retrive the arguments:
    pam = args['pam']
    rep_to_aaseq = args['rep_to_aaseq']
    acc_to_suffix = args['acc_to_suffix']
    cluster_to_rep = args['cluster_to_rep']
    
    
    # get the basename without extension:
    basename = os.path.basename(genome)
    accession, _ = os.path.splitext(basename)
    
    
    # create a query file for each genome: 
    sr_list = []
    with open(f'working/rec_overlap/queries/{accession}.query.faa', 'w') as w_handler: 
        for cluster in pam.index:
            cell = pam.loc[cluster, accession]
            if type(cell) == float:  # include only empty clusters
                rep = cluster_to_rep[cluster]
                seq = Seq.Seq(rep_to_aaseq[rep])
                sr = SeqRecord.SeqRecord(seq, id=cluster, description='')
                sr_list.append(sr) 
        count = SeqIO.write(sr_list, w_handler, "fasta")
    
    
    # create a blast database for the genome:
    os.makedirs(f'working/rec_overlap/databases/{accession}/', exist_ok=True)
    shutil.copyfile(genome, f'working/rec_overlap/databases/{accession}/{accession}.fna')  # just the content, not the permissions.
    command = f"""makeblastdb -in working/rec_overlap/databases/{accession}/{accession}.fna -dbtype nucl -parse_seqids"""  # '-parse_seqids' is required for 'blastdbcmd'.
    process = subprocess.Popen(command, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    process.wait()
    
    
    # perform the blast search:
    command = f'''tblastn \
        -query working/rec_overlap/queries/{accession}.query.faa \
        -db working/rec_overlap/databases/{accession}/{accession}.fna \
        -out working/rec_overlap/alignments/{accession}.tsv \
        -outfmt "6 {get_blast_header()}"
    '''
    process = subprocess.Popen(command, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    process.wait()
    
    
    # read the masked alignment: 
    colnames = f'{get_blast_header()}'.split(' ')
    masked = pnd.read_csv(f'working/rec_masking/alignments/{accession}.tsv', sep='\t', names=colnames )
    masked['qcov'] = round((masked['qend'] -  masked['qstart'] +1)/ masked['qlen'] * 100, 1)
    masked['scov'] = round((masked['send'] -  masked['sstart'] +1)/ masked['slen'] * 100, 1)
    
    
    # read the NON-masked alignment: 
    colnames = f'{get_blast_header()}'.split(' ')
    not_masked = pnd.read_csv(f'working/rec_overlap/alignments/{accession}.tsv', sep='\t', names=colnames )
    not_masked['qcov'] = round((not_masked['qend'] -  not_masked['qstart'] +1)/ not_masked['qlen'] * 100, 1)
    not_masked['scov'] = round((not_masked['send'] -  not_masked['sstart'] +1)/ not_masked['slen'] * 100, 1)
    

    # focus on genes that are still missing: 
    still_missing = set(not_masked['qseqid'].to_list())
    masked = masked[masked['qseqid'].isin(still_missing)]


    # masked: filter using the same thresholds of cd-hit:
    masked = masked[(masked['pident'] >= 90) & (masked['qcov'] < 70)]
    #masked = masked[masked['evalue'] <= 1e-5]  # if commented, more genes will be recovered
    masked = masked.sort_values('evalue', ascending=True)  # sort by evalue
    masked = masked.reset_index(drop=True)
    
    
    # NON-masked: filter using the same thresholds of cd-hit:
    not_masked = not_masked[(not_masked['pident'] >= 90) & (not_masked['qcov'] >= 70)]
    not_masked = not_masked[not_masked['evalue'] <= 1e-5]
    not_masked = not_masked.sort_values('evalue', ascending=True)
    not_masked = not_masked.reset_index(drop=True)


    # search for sequences that were partially recovered on rec_masked(), 
    # and that now they can be extended to completion. 
    improvements_df = []  # future dataframe showing the elongation to completion. 
    
    
    # group hsps by cluster: 
    groups_not_masked = not_masked.groupby('qseqid').groups
    groups_masked = masked.groupby('qseqid').groups
    for cluster in groups_not_masked.keys():
        if cluster not in groups_masked.keys():
            continue  # impossible to extend
        hpss_not_masked = not_masked.iloc[ groups_not_masked[cluster], ]
        hpss_masked = masked.iloc[ groups_masked[cluster], ]


        # keep the best hsp for this cluster (according to the evalue)
        hpss_masked = hpss_masked.drop_duplicates(subset='qseqid', keep='first')
        hpss_masked = hpss_masked.reset_index(drop=True)
        
        
        # get the contig where to look for a sequence elongation, plus the coverage to beat:
        contig_to_inspect = hpss_masked.loc[0, 'sseqid']
        cov_to_beat = hpss_masked.loc[0, 'qcovhsp']
        
        
        # search for candidates on that contig, having an higher converage:
        hpss_not_masked = hpss_not_masked[hpss_not_masked['sseqid'] == contig_to_inspect]
        hpss_not_masked = hpss_not_masked[hpss_not_masked['qcovhsp'] > cov_to_beat]
        if len(hpss_not_masked)==0: 
            continue  # no candidates found.
        # focus on the best candidated (according to the evalue):
        hpss_not_masked = hpss_not_masked.drop_duplicates(subset='qseqid', keep='first')  

        
        # save the elongation as a two-rows table: 
        elongation_df = pnd.concat([hpss_not_masked, hpss_masked], axis=0)
        elongation_df = elongation_df.reset_index(drop=True)
        elongation_df.loc[1, 'sseqid'] = elongation_df.loc[1, 'sseqid'] + ' (masked)'

        
        # populate the final improvements dataframe
        improvements_df.append(elongation_df)


    # get and save the final improvements dataframe
    if improvements_df != []:  # if something was recovered
        improvements_df = pnd.concat(improvements_df, axis=0)
        improvements_df = improvements_df.reset_index(drop=True)
    else:  # create an empty dataframe with compatible columns:
        # not_masked and masked should have the same header
        improvements_df = pnd.DataFrame(columns=not_masked.columns)
    improvements_df.to_csv(f'working/rec_overlap/elongations/{accession}.csv')


    # instantiate key objects: 
    cnt_good_genes = 0
    new_rows = []  # new rows for the sequences_df
    df_result = [] # using the common formatting
    
    
    # parse just the elongated sequences: 
    improvements_df = improvements_df[improvements_df['sseqid'].str.contains(' (masked)', regex=False) == False]
    for index, row in improvements_df.iterrows():
        cnt_good_genes += 1
        overlap_gid = f'{acc_to_suffix[accession]}_overlap_{cnt_good_genes}'


        # retrieve the sequence from the genome:
        start = int(row['sstart'])
        end = int(row['send'])
        contig = row["sseqid"]
        contig = contig.split('|')[1]   #split() because blast but some obscure formatting, eg: gb|NIGV01000003.1| for NIGV01000003.1.
        strand = '+'
        if start > end: # if on the other strand, invert the positions. 
            strand = '-'
            start, end = end, start
        curr_seq_translated, curr_seq_translated_tostop = \
            extract_aa_seq_from_genome(
                f'working/rec_overlap/databases/{accession}/{accession}.fna',
                contig, strand, start, end) 


        # if premature stop, sign the ID
        if len(curr_seq_translated_tostop) / len(curr_seq_translated) < 0.95 :
            overlap_gid = overlap_gid + '_stop'


        # populate the results dataframe:
        df_result.append({
            'ID': overlap_gid, 'cluster': cluster, 
            'accession': accession, 'contig': contig, 'strand': strand, 
            'start': start, 'end': end,
        })
        
        
        # populate the sequences dataframe
        new_rows.append({'cds': overlap_gid, 'accession': accession, 'aaseq': str(curr_seq_translated)})


    # save results for this genome:
    df_result = pnd.DataFrame.from_records(df_result)
    df_result.to_csv(f'working/rec_overlap/results/{accession}.csv')
    
    
    # return new rows for the sequences_df
    return new_rows



def recovery_overlap(logger, cores):
    
    
    # some log messages:
    logger.info("Recovering the genes checking the overlap...")
    
    
    # create sub-directories without overwriting:
    os.makedirs('working/rec_overlap/', exist_ok=True)
    os.makedirs('working/rec_overlap/queries/', exist_ok=True)
    os.makedirs('working/rec_overlap/databases/', exist_ok=True)
    os.makedirs('working/rec_overlap/alignments/', exist_ok=True)
    os.makedirs('working/rec_overlap/elongations/', exist_ok=True)
    os.makedirs('working/rec_overlap/results/', exist_ok=True)
    
    
    # check if it's everything pre-computed
    response = check_cached(
        logger, pam_path='working/rec_overlap/pam.csv',
        summary_path='working/rec_overlap/summary.csv',
        imp_files = [
            'working/rec_overlap/sequences.csv',
            'working/rec_overlap/seq_to_coords.pickle',])
    if response == 0: 
        return 0
    
    
    # load the assets to form the args dictionary:
    pam = pnd.read_csv('working/rec_masking/pam.csv', index_col=0)
    with open('working/clustering/acc_to_suffix.pickle', 'rb') as handler:
        acc_to_suffix = pickle.load(handler)
    with open('working/clustering/cluster_to_rep.pickle', 'rb') as handler:
        cluster_to_rep = pickle.load(handler)
    with open('working/clustering/rep_to_aaseq.pickle', 'rb') as handler:
        rep_to_aaseq = pickle.load(handler)
        
        
    # load the previously created species_to_genome: 
    with open('working/genomes/species_to_genome.pickle', 'rb') as handler:
        species_to_genome = pickle.load(handler)
    
    
    # create items for parallelization: 
    items = []
    for species in species_to_genome.keys(): 
        for genome in species_to_genome[species]: 
            items.append(genome)
    

    # randomize and divide in chunks: 
    chunks = chunkize_items(items, cores)
    
    
    # initialize the globalpool:
    globalpool = multiprocessing.Pool(processes=cores, maxtasksperchild=1)
    
    
    # start the multiprocessing: 
    results = globalpool.imap(
        load_the_worker, 
        zip(chunks, 
            range(cores), 
            itertools.repeat(['cds', 'accession', 'aaseq']), 
            itertools.repeat('cds'), 
            itertools.repeat(logger), 
            itertools.repeat(task_recoverlap),  # will return a new sequences dataframe (to be concat).
            itertools.repeat({'pam': pam, 'rep_to_aaseq': rep_to_aaseq, 'acc_to_suffix': acc_to_suffix, 'cluster_to_rep': cluster_to_rep}),
        ), chunksize = 1)
    all_df_combined = gather_results(results)
    
    
    # save tabular results:
    sequences_df = pnd.read_csv('working/rec_masking/sequences.csv', index_col=0)
    sequences_df_updated = pnd.concat([sequences_df, all_df_combined], axis=0)
    sequences_df_updated.to_csv('working/rec_overlap/sequences.csv')
    
    
    # empty the globalpool
    globalpool.close() # prevent the addition of new tasks.
    globalpool.join() 
    
    
    # update the pam:
    update_pam(logger, module_dir='working/rec_overlap', pam=pam)
    
    
    # update the seq to coordinates dictionary
    update_seq_to_coords(logger)
    
    
    # create the summary:
    create_summary(logger, module_dir='working/rec_overlap')
    
    
    return 0


import pickle
import cobra
from importlib import resources
import os


from ..commons import get_md5_string


from gempipe.curate.sanity import search_biomass


__PIMPCACHE__ = None


def get_annotation_info(model):
    
    
    # count how many annotation database for each object type (mets, reacs):
    # for metabolites: 
    m_annotations = set()
    for m in model.metabolites: 
        for a in m.annotation.keys():
            m_annotations.add(a)
    m_len = len(m_annotations)
    m_annotations = sorted(m_annotations)
    m_annotations= ', '.join(m_annotations)
    print(f"M ({m_len}):   " + str(m_annotations))
    
    
    # for reactions: 
    r_annotations = set()
    for r in model.reactions: 
        for a in r.annotation.keys():
            r_annotations.add(a)
    r_len = len(r_annotations)
    r_annotations = sorted(r_annotations)
    r_annotations= ', '.join(r_annotations)
    print(f"R ({r_len}):   " + str(r_annotations))
    
       
    
def make_memote_compliant(something_to_others):

                
    # make small adjustment to match the Memote/Miriam requirements:           
    for bigg_id in something_to_others.keys():
        for db in something_to_others[bigg_id].keys():
            
            
            if db =='inchikey':
                newlist = []
                for i in something_to_others[bigg_id][db]:
                    newlist.append(i.replace("InChIKey=", ''))
                something_to_others[bigg_id][db] = newlist
                
                
            if db =='chebi':
                newlist = []
                for i in something_to_others[bigg_id][db]:
                    newlist.append('CHEBI:' + i)
                something_to_others[bigg_id][db] = newlist
                
                
            if db == 'hmdb':
                # HMDB codes, to be accepted by Memote (MIRIAM) , must be of 5 
                # numbers, not 7. Comapare eg the following 'HMDB0001039', 'HMDB01039'.
                newlist = []
                for i in something_to_others[bigg_id][db]:
                    if len(i.replace('HMDB', ''))== 5:
                        newlist.append(i)
                something_to_others[bigg_id][db] = newlist
                
                
            if db == 'bigg.reaction':
                # avoid annotation repetition (eg UDCPDPS and R_UDCPDPS).
                newlist = []
                for i in something_to_others[bigg_id][db]:
                    if i.startswith('R_')==False:
                        newlist.append(i)
                something_to_others[bigg_id][db] = newlist
                
                
            if db == 'ec-code':
                # Reactions can have multiple EC codes. 
                # Not allowed are EC 1) incomplete or 2) containing letters.
                # Eg here we jeep just the second: 2.5.1.M1, 2.5.1.31, 2.5.1
                newlist = []
                for i in something_to_others[bigg_id][db]:
                    if len(i.split('.'))==4 and i.replace('.', '').isnumeric():
                        newlist.append(i)
                something_to_others[bigg_id][db] = newlist
                
                
            if db == 'brenda':
                # exaclty like the EC codes
                newlist = []
                for i in something_to_others[bigg_id][db]:
                    if len(i.split('.'))==4 and i.replace('.', '').isnumeric():
                        newlist.append(i)
                something_to_others[bigg_id][db] = newlist
                
                
            if db == 'pubchem.compound':
                newlist = []
                for i in something_to_others[bigg_id][db]:
                    newlist.append(i.replace("CID", ''))  # eg from 'CID135398555' to '135398555'
                something_to_others[bigg_id][db] = newlist
                
                
            if db == 'reactome':
                newlist = []
                for i in something_to_others[bigg_id][db]:
                    newlist.append(i.rsplit('.', 1)[0])  # eg from 'R-HSA-70467.7' to 'R-HSA-70467'.
                something_to_others[bigg_id][db] = newlist
                
    
    return something_to_others
    

    
def boost_annotations(logger, x, something_to_others, mrswitch='r', overwrite=False): 
    # Take a single reaction or metabolite and improve its annotations based on prebuilt mnx dicts.
    
    
    #  Works both for metabolites and reactions.
    if mrswitch=='r':  
        mrid = x.id
        # correct for ModelSEED reaction ids: 
        if mrid.startswith('rxn') and mrid[-3]=='_':  # for example "rxn11567_c0"
            mrid = mrid[ : -3]
    else: mrid = x.id.rsplit('_', 1)[0]  # remove compartment

    
    # extract all the annots provided by mnx for this ID:
    cnt = 0
    if mrid in something_to_others.keys(): 
        full_annots = something_to_others[mrid]

        
        # format model's annotations as set: 
        if not overwrite: 
            for annot in x.annotation.keys():
                if type(x.annotation[annot]) == str: 
                    x.annotation[annot] = set([x.annotation[annot]])
                elif type(x.annotation[annot]) == list: 
                    x.annotation[annot] = set(x.annotation[annot])
                else: 
                    logger.error(f"Found strange annotation type ({str(type(x.annotation[annot]))}) for this annotation: {x.annotation[annot]}. Please contact the developer.")
                    return 1
        else: # the user requested an overwriting
            for annot in x.annotation.keys():
                x.annotation[annot] = set()

        
        # iterate through variuos databases provided by metanetx: 
        for db in full_annots.keys():
            # no annotations provided for this database:
            if db not in x.annotation.keys():
                x.annotation[db] = set()
            # copy annotations: 
            for annot in full_annots[db]:
                if annot not in x.annotation[db]:
                    x.annotation[db].add(annot)
                    cnt += 1  # 1 new annotation added !


        # re-format annotations as lists:
        for db in x.annotation.keys():
            x.annotation[db] = list(x.annotation[db]) 
            

    return x.annotation, cnt



def write_generic_sboterms(logger, model, overwrite=False):
    
    cnt = 0
    
    def annotate_item(logger, item, mode, cnt, biom_ids=None):
        
        # PART A)
        # create the new term if not present:
        if 'sbo' not in item.annotation.keys():
            item.annotation['sbo'] = []
              
        # PART B)
        # format model's annotations as set: 
        if not overwrite: 
            if type(item.annotation['sbo']) == str: 
                item.annotation['sbo'] = set([item.annotation['sbo']])
            elif type(item.annotation['sbo']) == list: 
                item.annotation['sbo'] = set(item.annotation['sbo'])
            else: 
                logger.error(f"Found strange annotation type ({str(type(item.annotation['sbo']))}) for this annotation: {item.annotation['sbo']}. Please contact the developer.")
                return None
        else: # the user requested an overwriting
            item.annotation['sbo'] = set()
            
        # PART C)
        # add new SBO annotations.
        # all SBO terms can be viewed at https://www.ebi.ac.uk/ols4/ontologies/sbo.
        if mode=='g':
            item.annotation['sbo'].add('SBO:0000243')  # generic gene
            cnt += 1  # 1 new annotation added !
        if mode=='r':
            if len(item.metabolites)== 1:  # exchange/sink/demand
                if list(item.metabolites)[0].id.rsplit('_', 1)[-1] == 'e':  # exchange
                    item.annotation['sbo'].add('SBO:0000627')  # generic exchange reaction
                    cnt += 1  # 1 new annotation added !
                elif item.lower_bound < 0 and item.upper_bound > 0: # sink
                    item.annotation['sbo'].add('SBO:0000632')  # generic sink reaction
                    cnt += 1  # 1 new annotation added !
                else : # demand
                    item.annotation['sbo'].add('SBO:0000628')  # generic demand reaction
                    cnt += 1  # 1 new annotation added !
            else:  # metabolic or transport or biomass
                # get the compartments involved
                comps_involved = set([m.id.rsplit('_', 1)[-1] for m in item.metabolites])
                if item.id in biom_ids:  # biomass
                    item.annotation['sbo'].add('SBO:0000629')  # generic biomass
                    cnt += 1  # 1 new annotation added !
                elif len(comps_involved) > 1:  # transport
                    item.annotation['sbo'].add('SBO:0000655')  # generic transport
                    cnt += 1  # 1 new annotation added !
                else:  # pure metabolic
                    item.annotation['sbo'].add('SBO:0000176')  # generic metabolic
                    cnt += 1  # 1 new annotation added !
        if mode=='m':
            item.annotation['sbo'].add('SBO:0000247')  # generic metabolite
            cnt += 1  # 1 new annotation added !
        

        # PART D)
        # re-format annotations as lists:
        item.annotation['sbo'] = list(item.annotation['sbo'])
        
        return cnt
    
    
    # SBO for genes:
    for g in model.genes:
        cnt = annotate_item(logger, g, 'g', cnt)
    # SBO for reactions:
    biom_ids = search_biomass(model, verbose=False)  # get biomass reaction IDs.
    for r in model.reactions:
        cnt = annotate_item(logger, r, 'r', cnt, biom_ids)
    # SBO for metabolites:
    for m in model.metabolites:
        cnt = annotate_item(logger, m, 'm', cnt)
        

    return cnt
    
    
    
    

def pimp_my_model(logger, model, fromchilds=False, overwrite=False):
    
    
    # autodetect ID system looking for water in cytosol
    id_sys = 'bigg'
    """
    id_sys = None
    for m in model.metabolites:
        # remove cytosol annot
        mid = m.id.rsplit('_', 1)[0]
        if mid == 'h2o': 
            id_sys = 'bigg'
            break
        elif mid == 'cpd00001':
            id_sys = 'seed'
            break
    if verbose: print(f'"{id_sys}" detected.', flush=True)
    if id_sys == None: id_sys = 'bigg'
    """
    
    
    # load the prebuilt dictionary only if this was the first request: 
    global __PIMPCACHE__
    if __PIMPCACHE__ == None:
        __PIMPCACHE__ = {} # flag for 'already loaded'
        logger.debug("Loading MetaNetX prebuilt dicts... ")
        
        
        if not fromchilds:
            with resources.path("gempipe.assets", f"{id_sys}_to_others_M.pickle") as asset_path: 
                with open(asset_path, 'rb') as handler:
                    __PIMPCACHE__['something_to_others_M'] = pickle.load(handler)
            with resources.path("gempipe.assets", f"{id_sys}_to_others_R.pickle") as asset_path: 
                with open(asset_path, 'rb') as handler:
                    __PIMPCACHE__['something_to_others_R'] = pickle.load(handler)


        else:  # required 'child' annotations: 
            with resources.path("gempipe.assets", f"{id_sys}_to_others_extended_M.pickle") as asset_path: 
                with open(asset_path, 'rb') as handler:
                    __PIMPCACHE__['something_to_others_M'] = pickle.load(handler)
            with resources.path("gempipe.assets", f"{id_sys}_to_others_extended_R.pickle") as asset_path: 
                with open(asset_path, 'rb') as handler:
                    __PIMPCACHE__['something_to_others_R'] = pickle.load(handler)
                    
    
    # get what has been loaded:
    something_to_others_M = __PIMPCACHE__['something_to_others_M']
    something_to_others_R = __PIMPCACHE__['something_to_others_R']
        
    
    # apply small corrections to meet the Memote/Miriam standards:
    something_to_others_M = make_memote_compliant(something_to_others_M)
    something_to_others_R = make_memote_compliant(something_to_others_R)
                    
    
    logger.debug("Annotating metabolites...")   
    m_cnt = 0  # counter for logger
    for m in model.metabolites:
        response = boost_annotations(logger, m, something_to_others_M, 'm', overwrite)
        if response == 1: return 1
        else: m.annotation, cnt = response
        m_cnt = m_cnt + cnt
    logger.debug(f'{m_cnt} annots added.')
    
    
    logger.debug("Annotating reactions... ")
    r_cnt = 0  # counter for logger
    for r in model.reactions:
        response = boost_annotations(logger, r, something_to_others_R, 'r', overwrite)
        if response == 1: return 1
        else: r.annotation, cnt = response
        r_cnt = r_cnt + cnt
    logger.debug(f'{r_cnt} annots added.')  
    
    
    logger.debug("Annotating SBO terms... ")
    sbo_cnt = write_generic_sboterms(logger, model, overwrite)
    if sbo_cnt == None: return 1
    logger.debug(f'{sbo_cnt} SBO terms added.')  
    
    
    return 0
        
        

def denovo_annotation(logger):
    
    
    # log some message: 
    logger.info("Performing de-novo model annotation...")

    
    # check presence of already computed files 
    if os.path.exists(f'working/duplicates/draft_panmodel.json') and os.path.exists(f'working/duplicates/md5.pickle'):
        if os.path.exists(f'working/duplicates/draft_panmodel_da.json') and os.path.exists(f'working/duplicates/md5_da.pickle'):
            with open('working/duplicates/md5.pickle', 'rb') as handler:
                md5 = pickle.load(handler)
            with open('working/duplicates/md5_da.pickle', 'rb') as handler:
                md5_da = pickle.load(handler)
            # compare md5:
            if md5 == md5_da == get_md5_string('working/duplicates/draft_panmodel.json'):
                # log some message: 
                logger.info('Found all the needed files already computed. Skipping this step.')
                # signal to skip this module:
                return 0
    
    
    # load the final draft panmodel
    draft_panmodel = cobra.io.load_json_model('working/duplicates/draft_panmodel.json')
    
                
    # denovo annotation (M/R/SBO)
    response = pimp_my_model(logger, draft_panmodel, overwrite=True)
    if response == 1: return 1

    
    # replace file on disk:
    cobra.io.save_json_model(draft_panmodel, 'working/duplicates/draft_panmodel_da.json')
    
    
    # trace the parent model md5:
    parent_md5 = get_md5_string('working/duplicates/draft_panmodel.json')
    with open('working/duplicates/md5_da.pickle', 'wb') as handle:
        pickle.dump(parent_md5, handle)
    
    
    return 0 
    
    
    


import os
import pickle
import subprocess


import pandas as pnd
from Bio import SeqIO, SeqRecord, Seq


from ..commons import get_outdir



def func_annot(logger, cores, outdir, dbs, dbmem): 
    
    
    # create subdirs without overwriting
    os.makedirs('working/annotation/', exist_ok=True)
    dbs = get_outdir(dbs)  # append a '/' and create if necessary.
    
 
    
    # print some log messages
    logger.info('Performing functional annotation of the representative sequences...')
    
    
    # load the final PAM: 
    pam = pnd.read_csv(outdir + 'pam.csv', index_col=0)
    
    
    # get the accessions retained:
    accessions = set()
    with open('working/proteomes/species_to_proteome.pickle', 'rb') as handler:
        species_to_proteome = pickle.load(handler)
        for species in species_to_proteome.keys(): 
            for proteome in species_to_proteome[species]:
                basename = os.path.basename(proteome)
                accession, _ = os.path.splitext(basename)
                accessions.add(accession)
    
    
    # check if all the output where already computed:
    if os.path.exists('working/annotation/proc_acc.pickle'):
        with open('working/annotation/proc_acc.pickle', 'rb') as handler:
            proc_acc = pickle.load(handler) 
        if accessions == proc_acc == set(list(pam.columns)):
            if os.path.exists('working/annotation/pan.emapper.annotations'):
                if os.path.exists('working/annotation/representatives.faa'):
                    seq_ids = [] 
                    with open('working/annotation/representatives.faa', 'r') as r_handler:
                        for seqrecord in SeqIO.parse(r_handler, "fasta"):
                            seq_ids.append(seqrecord.id)
                    clusters = list(pam.index)
                    if set(seq_ids) == set(clusters):
                        # log some message: 
                        logger.info('Found all the needed files already computed. Skipping this step.')
                        # signal to skip this module:
                        return 0
    
    
    # load the sequences resources: 
    with open('working/clustering/cluster_to_rep.pickle', 'rb') as handler:
        cluster_to_rep = pickle.load(handler)
    with open('working/clustering/rep_to_aaseq.pickle', 'rb') as handler:
        rep_to_aaseq = pickle.load(handler)
        
        
    # parse the pam to create a single input fasta files with representative sequences: 
    sr_list = []
    for cluster in pam.index:
        rep = cluster_to_rep[cluster]
        aaseq = Seq.Seq(rep_to_aaseq[rep])
        sr = SeqRecord.SeqRecord(aaseq, id=cluster, description=f'({rep})')
        sr_list.append(sr)
    with open(f'working/annotation/representatives.faa', 'w') as w_handler:
        count = SeqIO.write(sr_list, w_handler, "fasta")
        
        
    # check if the database already exists:
    if (not os.path.exists(dbs + 'eggnog_proteins.dmnd')) or (not os.path.exists(dbs + 'eggnog.db')):
        logger.info("The database for functional annotation is missing. It will be dowloaded now...")
        with open(f'working/logs/stdout_funcdownload.txt', 'w') as stdout, open(f'working/logs/stderr_funcdownload.txt', 'w') as stderr: 
            command = f"""download_eggnog_data.py -y --data_dir {dbs}"""
            process = subprocess.Popen(command, shell=True, stdout=stdout, stderr=stderr)
            process.wait()
        logger.info("Download completed. Now executing the annotation...")
        
    
    # execute the command.
    # --dbmem loads the whole eggnog.db sqlite3 annotation database during the annotation step, and 
    # therefore requires ~44 GB of memory. It is recommanded when annotating a large number of sequences.
    # download_eggnog_data.py : This will download the eggNOG annotation database (along with the taxa databases), 
    # and the database of eggNOG proteins for Diamond searches.
    with open(f'working/logs/stdout_funcannot.txt', 'w') as stdout, open(f'working/logs/stderr_funcannot.txt', 'w') as stderr: 
        command = f"""emapper.py \
            --cpu {cores} \
            --override \
            --data_dir {dbs} \
            -i working/annotation/representatives.faa \
            -m diamond \
            --itype proteins \
            --trans_table 11 \
            --excel {'--dbmem' if dbmem else ''} \
            --output pan; mv pan.* working/annotation/"""
        process = subprocess.Popen(command, shell=True, stdout=stdout, stderr=stderr)
        process.wait()

        
    # make traces to keep track of the accessions processed
    proc_acc = set(list(pam.columns))
    with open('working/annotation/proc_acc.pickle', 'wb') as handler:
        pickle.dump(proc_acc, handler)
    
    
    return 0


import os
import glob
import subprocess
import pickle


import pandas as pnd
from Bio import SeqIO, SeqRecord, Seq



def get_sequences_and_source(gbff_file, ):

    # use SeqIO.read() if you expect only one record in the GenBank file.
    # alternatively, use SeqIO.parse() if there are multiple records in the file.
    records = SeqIO.parse(gbff_file, 'genbank')
                              
    sequences = []
    sourcedb = None

    for record in records:

        sequence_id = record.id
        sequence_description = record.description
        sequence_seqlen = len(record.seq)
        
        
        # append sequence id
        sequences.append(sequence_id)
        
        
    # take the first and search for an underscore (_). 
    # if found, we assume the gbff is from refseq, not genbank.
    source_db = 'refseq' if '_' in sequences[0] else 'genbank'
                              
    return sequences, source_db



def get_uniprot_table(accession, sequences, source_db, rformat):
    
    
    # choose the crossref db (embl can be seen as a replacement for genbank)
    if source_db=='genbank': 
        crossref = 'embl'
    else: crossref = 'refseq'
    
    
    # adjust sequences if from genbank: 
    if crossref=='embl':
        sequences = [i.rsplit('.', 1)[0] for i in sequences]


    # define the query string based on the format:
    query_string = '%28' + '+OR+'.join([f'%28xref%3A{crossref}-{i}%29' for i in sequences]) + '%29'

    if rformat == 'json':  
        # 'json' parsing still to implement. 'tsv' preferred because it weights much less.
        query_string = f'https://rest.uniprot.org/uniprotkb/stream?format={rformat}&query={query_string}'

    if rformat == 'tsv':  
        query_fields = ['accession', 'gene_oln', 'xref_refseq', 'xref_embl', 'xref_kegg', 'xref_geneid',]
        query_fields = '%2C'.join(query_fields) 
        query_string = f'https://rest.uniprot.org/uniprotkb/stream?fields={query_fields}&format={rformat}&query={query_string}'


    # define the tmp output text file:  
    outpath = f'working/gannots/response_{accession}.tsv'

    
    # perform the query: 
    command = f'''wget -O {outpath} --quiet "{query_string}"'''
    process = subprocess.Popen(command, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    process.wait()


    response = pnd.read_csv(outpath, sep='\t')
    os.remove(outpath)  # clean up, table is in ram.
    return response
    
    
    
def get_features(accession, gbff_file):
    
    
    # use SeqIO.read() if you expect only one record in the GenBank file.
    # alternatively, use SeqIO.parse() if there are multiple records in the file.
    records = SeqIO.parse(gbff_file, 'genbank')
                              

    results_df = []
    for record in records:

        sequence_id = record.id
        sequence_description = record.description
        sequence_seqlen = len(record.seq)
        
        
        # access features:
        for feature in record.features:
            if feature.type == 'CDS':
                try: location = feature.location
                except: location = None
                try: gene_symbol = feature.qualifiers['gene'][0]
                except: gene_symbol = None
                try: aa_seq = feature.qualifiers['translation'][0]
                except: 
                    try: aa_seq = str(feature.extract(record.seq).translate())  # reverse complement automatically handled
                    except: aa_seq = None
                try: locus_tag = feature.qualifiers['locus_tag'][0]
                except: locus_tag = None
                try: old_locus_tag = feature.qualifiers['old_locus_tag'][0]
                except: old_locus_tag = None
                if locus_tag != None and (type(locus_tag)==int or str(locus_tag).isdigit()):  
                    # gene nume is just a number (the genbank may come from prodigal/bactabolize like pipelines)
                    # probably the gene ID will conflict with genes from other strains. 
                    # Therefore we prepend its accession:
                    locus_tag = accession + '_' + record.id + '_' + str(locus_tag)
                    locus_tag = locus_tag.replace('.', '').replace('-', '')  # remove undesired chars
                    
                # Note on the 'protein_id' field: 
                # if the .gbff is from GenBank, then 'protein_id' is the actual protein (example: CCC77580.1)
                # if the .gbff is from RefSeq, then 'protein_id' is a NR protein (prefix WP_)
                try: protein_id = feature.qualifiers['protein_id'][0]
                except: protein_id = None
                
                results_df.append({
                    'accession': accession, 'seqid': record.id, 'gene_symbol': gene_symbol, # 'location': location,  
                    'locus_tag': locus_tag, 'old_locus_tag': old_locus_tag, 'protein_id': protein_id})
    
    
    results_df = pnd.DataFrame.from_records(results_df)
    return results_df
    
    
    
def get_gannots_tabular(accession, gbff_file, outfolder):
    
    
    # get accession list and source database (genbank/refseq):
    sequences, source_db = get_sequences_and_source(gbff_file, )
    
    # get the uniprot crossref table:
    up_table = get_uniprot_table(accession, sequences, source_db, 'tsv')
    
    # parse gain the gbff obtaining the features (tabular):
    feat_table = get_features(accession, gbff_file)

        
    # decide between 'locus_tag' and 'old_locus_tag' using intersection:
    i_locus     = set(feat_table['locus_tag'    ].to_list()).intersection(set(up_table['Gene Names (ordered locus)'].to_list()))
    i_old_locus = set(feat_table['old_locus_tag'].to_list()).intersection(set(up_table['Gene Names (ordered locus)'].to_list()))
    
    if len(i_old_locus) > len(i_locus):
        left_on = 'old_locus_tag'
    else: left_on = 'locus_tag'
    right_on = 'Gene Names (ordered locus)'
      
        
    if len(up_table)!=0 and (len(i_locus)>0 or len(i_old_locus)>0):  # at least one seq available in uniprot
    
        # merge based on different column names
        merge_table = pnd.merge(feat_table, up_table, how='inner', left_on=left_on, right_on=right_on)
        
        # select matching IDs:
        merge_table = merge_table[merge_table[left_on].notna()]
        # explanations on NCBI protein IDs: https://www.ncbi.nlm.nih.gov/refseq/about/prokaryotes/reannotation/
        # if the .gbff is from GenBank, then 'protein_id' is the actual protein (example: CCC77580.1)
        # if the .gbff is from RefSeq, then 'protein_id' is a NR protein (prefix WP_)
        columns_to_keep = ['accession', 'seqid', 'gene_symbol', left_on, 'protein_id', 'Entry', 'RefSeq', 'KEGG', 'GeneID']
        merge_table = merge_table[columns_to_keep]
        
        
    else:  # uniprot data not available: 
        merge_table = feat_table.copy()
        
        columns_to_keep = ['accession', 'seqid', 'gene_symbol', left_on, 'protein_id', ]
        merge_table = merge_table[columns_to_keep]
        
        # add missing columns: 
        for i in ['Entry', 'RefSeq', 'KEGG', 'GeneID']:
            merge_table[i] = None
        
        
    
    # rename, reorder and select columns: 
    if source_db == 'genbank':
        renamed_cols = ['accession', 'seqid', 'gene_symbol', 'gid', 'ncbiprotein', 'uniprot', 'refseq', 'kegg', 'ncbigene']
    else:  # source_db == 'refseq'
        merge_table = merge_table.drop(columns='RefSeq')   # in this case it would be redundant
        # in case of 'refseq', the crossref 'ncbiprotein' is not available 
        merge_table['ncbiprotein'] = None
        renamed_cols = ['accession', 'seqid', 'gene_symbol', 'gid', 'refseq', 'uniprot', 'kegg', 'ncbigene', 'ncbiprotein']  
    # finally rename columns
    merge_table = merge_table.rename(columns=dict(zip(merge_table.columns, renamed_cols)))
    # reorder and select columns:
    merge_table = merge_table[['accession', 'gid', 'refseq', 'ncbiprotein', 'ncbigene', 'kegg', 'uniprot' ]]
    
    
    
    # if source_db=='genbank' , then the 'WP_' accessions of 'refseq' must be taken from 'up_table'.
    if source_db=='genbank':
        formatted_col = []
        for i in merge_table['refseq']:
            wp_code = None   # default
            if type(i) == str:  # avoid empty cells: 
                for code in i.split(';'):
                    #if code.startswith('WP_'):  # Memote is not yet ready for the new format.
                    if any([code.startswith(i) for i in "AC|AP|NC|NG|NM|NP|NR|NT|NW|XM|XP|XR|YP|ZP".split('|')]):
                        wp_code = code
                        break
            formatted_col.append(wp_code)
        # replace with formatted version:
        merge_table['refseq'] = formatted_col
        
        
    # finally, the 'ncbigene' and 'kegg' columns need to be formatted:
    def simple_formatter(col):
        formatted_col = []
        for i in col: 
            formatted_cell = None
            if type(i) == str:  # avoid emtpy cells:
                formatted_cell = i.split(';')[0]
            formatted_col.append(formatted_cell)
        return formatted_col
    merge_table['ncbigene'] = simple_formatter(merge_table['ncbigene'])
    merge_table['kegg'] = simple_formatter(merge_table['kegg'])
    

    
    merge_table.to_csv(f'working/gannots/gannots_{accession}.csv')
    return merge_table



def create_proteome_from_gbff(accession, gbff_file, gannots, outfolder):
    
    
    # use SeqIO.read() if you expect only one record in the GenBank file.
    # alternatively, use SeqIO.parse() if there are multiple records in the file.
    records = SeqIO.parse(gbff_file, 'genbank')
                              

    sr_list = []
    for record in records:        
        
        # access features:
        for feature in record.features:
            if feature.type == 'CDS':
                try: location = feature.location
                except: location = None
                try: gene_symbol = feature.qualifiers['gene'][0]
                except: gene_symbol = None
                try: aa_seq = feature.qualifiers['translation'][0]
                except: 
                    try: aa_seq = str(feature.extract(record.seq).translate())  # reverse complement automatically handled
                    except: aa_seq = None
                try: locus_tag = feature.qualifiers['locus_tag'][0]
                except: locus_tag = None
                if locus_tag != None and (type(locus_tag)==int or str(locus_tag).isdigit()):  
                    # gene nume is just a number (the genbank may come from prodigal/bactabolize like pipelines)
                    # probably the gene ID will conflict with genes from other strains. 
                    # Therefore we prepend its accession:
                    locus_tag = accession + '_' + record.id + '_' + str(locus_tag)
                    locus_tag = locus_tag.replace('.', '').replace('-', '')  # remove undesired chars
                try: old_locus_tag = feature.qualifiers['old_locus_tag'][0]
                except: old_locus_tag = None
                try: protein_id = feature.qualifiers['protein_id'][0]
                except: protein_id = None
                
                
                # get the right 'gid'
                if old_locus_tag != None and old_locus_tag in gannots['gid'].to_list():
                    gid = old_locus_tag
                else: gid = locus_tag
                
                
                # append the seqrecord:
                if aa_seq != None and gid != None:
                    sr = SeqRecord.SeqRecord(Seq.Seq(aa_seq), id=gid, description=f'(file: {gbff_file} sequence: {record.id})')
                    sr_list.append(sr)
            
    
    # finally write the proteome in fasta format:
    proteome_filepath = outfolder + f'{accession}.faa'
    with open(proteome_filepath, 'w') as w_handler:
        count = SeqIO.write(sr_list, w_handler, "fasta")
        
        
    return proteome_filepath
        
                
        
def create_proteome_and_gannots(gbff_file, outfolder):
    
    
    # get the "accession" for this file (just the simple filename)
    basename = os.path.basename(gbff_file)
    accession, _ = os.path.splitext(basename)
    
                       
    # create gene annotation pnd.DataFrame (gannots):
    gannots = get_gannots_tabular(accession, gbff_file, outfolder)
    
    
    # create proteome fasta file:
    proteome_filepath = create_proteome_from_gbff(accession, gbff_file, gannots, outfolder)
    
    
    return proteome_filepath, gannots



def check_cached_parsing(logger, species_to_genbank, outdir): 
    
    
    # verify the presence of needed files, and return 
    # the 'species_to_proteome' if they are all available.
    species_to_proteome = {}
    
    
    # if present , load the 'all_gannots' dataframe: 
    if not os.path.exists(outdir + 'gannots.csv'):
        return False
    # "low_memory=False" to prevent "DtypeWarning: Columns ... have mixed types."
    all_gannots = pnd.read_csv(outdir + 'gannots.csv', index_col=0, low_memory=False)
    avail_accessions = set(all_gannots['accession'].to_list())
        
    
    # check the presence of proteomes and gene annotation files: 
    required_accessions = set()
    for species in species_to_genbank.keys(): 
        species_to_proteome[species] = []
        
        for file in species_to_genbank[species]:
            basename = os.path.basename(file)
            accession, _ = os.path.splitext(basename)
            required_accessions.add(accession)
            
            # return empty if not found
            if not os.path.exists('working/proteomes/' + accession + '.faa'):
                return False
            if not os.path.exists('working/gannots/gannots_' + accession + '.csv'):
                return False
            
            # populate the 'species_to_proteome' dict if files are already available:
            species_to_proteome[species].append('working/proteomes/' + accession + '.faa')
            
            
    # the 'all_gannots' dataframe must have exaclty the same accessions, no more, no less:
    if required_accessions != avail_accessions:
        return False
            
            
    # save the dictionary to disk: 
    with open('working/proteomes/species_to_proteome.pickle', 'wb') as file:
        pickle.dump(species_to_proteome, file)
           
            
    return True
            
    

def handle_manual_genbanks(logger, genbanks, outdir):
    
    
    # create a species-to-genome dictionary
    species_to_proteome = {}
    species_to_genbank = {}
    logger.debug(f"Checking the formatting of the provided -gb/--genbanks attribute...") 
    
    
    
    # PART A) populate the species_to_genbank dictionary.
    
    # check if the user specified a folder:
    if os.path.exists(genbanks):
        if os.path.isdir(genbanks):
            if genbanks[-1] != '/': genbanks = genbanks + '/'
            files = glob.glob(genbanks + '*')
            species_to_genbank['Spp'] = files
    
    elif '+' in genbanks and '@' in genbanks: 
        for species_block in genbanks.split('+'):
            species, files = species_block.split('@')
            for file in files.split(','): 
                if not os.path.exists(file):
                    logger.error("The following file provided in -gb/--genbanks does not exists: " + file)
                    return 1
            species_to_genbank[species] = files.split(',')
            
    else: # the user has just 1 species
        for file in genbanks.split(','): 
            if not os.path.exists(file):
                logger.error("The following file provided in -gb/--genbanks does not exists: " + file)
                return 1
        species_to_genbank['Spp'] = ganbanks.split(',')

    
    # report a summary of the parsing: 
    logger.info(f"Inputted {len(species_to_genbank.keys())} species with well-formatted paths to genbanks.") 
    
    

    # PART B) parse each genbank file, and populate the species_to_proteome dictionary.
    # Moreover, concat the gannots to create a single annotation table. 
    logger.info("Parsing genbanks file to extract proteomes and gene annotations...")
    os.makedirs('working/proteomes/', exist_ok=True)
    os.makedirs('working/gannots/', exist_ok=True)
    
    
    # check if files were already computed:
    if check_cached_parsing(logger, species_to_genbank, outdir):
        logger.info("Found all the needed files already computed. Skipping this step.")
        
        return 0
    
    
    # parse each genbank file:
    all_gannots = []
    for species in species_to_genbank.keys():
        created_files = []
        for file in species_to_genbank[species]:
            
            # all proteomes and gannots will be stored in 'working/proteomes/'.
            proteome_filepath, gannots = create_proteome_and_gannots(file, outfolder='working/proteomes/')
            created_files.append(proteome_filepath)
            all_gannots.append(gannots)
        species_to_proteome[species] = created_files
    logger.debug(f"Input genbanks converted to proteomes in ./working/proteomes/.")
    logger.debug(f"Created the species-to-proteome dictionary: {str(species_to_proteome)}.") 

    
    # create a single gene annotation table as final 'gempipe recon' output: 
    all_gannots = pnd.concat(all_gannots)
    all_gannots = all_gannots.astype(str)  # 'ncbigene' is still formatted as number
    all_gannots = all_gannots.replace('nan', None)
    #all_gannots['refseq'] = [i.rsplit('.', 1)[0] if type(i)==str else None for i in all_gannots['refseq'] ]
    all_gannots['ncbigene'] = [i.rsplit('.', 1)[0] if type(i)==str else None for i in all_gannots['ncbigene'] ]
        
    
    all_gannots.to_csv(outdir + 'gannots.csv')
    logger.debug(f"Saved the genes annotation table to {outdir + 'all_gannots.csv'}.")
    
    
    # save the dictionary to disk: 
    with open('working/proteomes/species_to_proteome.pickle', 'wb') as file:
        pickle.dump(species_to_proteome, file)
    logger.debug(f"Saved the species-to-proteome dictionary to file: ./working/proteomes/species_to_proteome.pickle.")
    
    
    return 0


import os
import pickle


import pandas as pnd
import cobra


from ..commons import get_md5_string
from .refexpansion import mancor_to_dict



def get_first_occurrence(puremid, model):
    
    
    # get the fisrt metabolite in a model matching the provided 'puremid'.
    # 'puremid' is a metabolite ID without compartment. 
    for m in model.metabolites: 
        if puremid == m.id.rsplit('_', 1)[0] :
            return m
    return None


    
def get_translation_dictionary_mnx(model, mrmode='m'): 

    
    # even MNX is not perfect, for example lald__L and abt__L are clearly different species.
    to_skip_m = ['MNXM732880']  # 'MNXM732880': {'lald__L', 'abt__L'}
    to_skip_r = []
    if   mrmode == 'm': to_skip = to_skip_m
    elif mrmode == 'r': to_skip = to_skip_r
    
    
    # Dictionary grouping duplicated m/r. Later a 'good' member will be determined for each group. 
    # Here each group of metabolites could have different f/c.
    mnx_to_mrids = {}  # future 'to_translate'
    
    
    # iterate through all m/r in this model
    if   mrmode == 'm': items = model.metabolites
    elif mrmode == 'r': items = model.reactions
    for item in items:

        
        # get the annotations
        try: # mnxannots can be list of str (if just 1 element)
            if   mrmode == 'm': mnxannots = item.annotation['metanetx.chemical']
            elif mrmode == 'r': mnxannots = item.annotation['metanetx.reaction']
            if type(mnxannots) == str: mnxannots = [mnxannots]
        except: continue
        
        
        # populate the dict
        for mnxannot in mnxannots: 
            if mnxannot in to_skip: 
                continue
            if mnxannot not in mnx_to_mrids.keys():
                mnx_to_mrids[mnxannot] = set()
            if   mrmode == 'm': mnx_to_mrids[mnxannot].add(item.id.rsplit('_', 1)[0])  # without compartment. 
            elif mrmode == 'r': 
                if item.id.startswith("EX_") and len(item.metabolites)==1: 
                    continue  # by design not interest in EX_ reactions
                mnx_to_mrids[mnxannot].add(item.id) 
            
            
    # filter to keep only groups with >1 elements.
    to_translate = {}
    for mnxannot in mnx_to_mrids.keys(): 
        if len(mnx_to_mrids[mnxannot]) >= 2:
            to_translate[mnxannot] = mnx_to_mrids[mnxannot]
                    
                    
    return to_translate

    

def sort_translation_dictionary(to_translate, model, refmodel, reffree, mrmode='m'):
    
    
    # 'to_translate' contains duplicated metabolites divided in groups. 
    # here, for each group, we select only one 'good' metabolite to maintain. 
    # Therefore, dictionary 1-to-many will be converted in a 1-to-1. 
    to_translate_11 = {}
    
    
    # get all mids without compartment
    if   mrmode == 'm': 
        mrids_refmodel = set([m.id.rsplit('_', 1)[0] for m in refmodel.metabolites])
        mrids_reffree = set([m.id.rsplit('_', 1)[0] for m in reffree.metabolites])
    elif mrmode == 'r': 
        mrids_refmodel = set([r.id for r in refmodel.reactions])
        mrids_reffree = set([r.id for r in reffree.reactions])
    
    
    results_df = []
    for group in to_translate.values():
        
        # determine the good metabolite to represent the group:
        good_mrid = list(group)[0]  # begin with the first
        for mrid in group:
            if mrid in mrids_refmodel: 
                good_mrid = mrid  # give precedence to refmodel
                break
                
                
        # solve in a 1-to-1 dict
        for mrid in group:
            if mrid != good_mrid: 
                # get formula, charge, and flags:
                
                
                # for metabolties:  
                if mrmode == 'm':
                    
                    # for this metabolite to replace
                    m = get_first_occurrence(mrid, model)
                    formula = m.formula
                    charge = m.charge
                    in_refmodel = mrid in mrids_refmodel
                    in_reffree = mrid in mrids_reffree

                    # for the replacement metabolite
                    good_m = get_first_occurrence(good_mrid, model)
                    good_formula = good_m.formula
                    good_charge = good_m.charge
                    good_in_refmodel = good_mrid in mrids_refmodel
                    good_in_reffree = good_mrid in mrids_reffree


                    # WARNING: by design, skip if both metabolite are ancoded in the refmodel:
                    if in_refmodel and good_in_refmodel: 
                        continue
                        

                    # log the future edits: 
                    new_row = {
                        'duplicated': mrid, 'd_formula': formula, 'd_charge': charge, 'd_in_refmodel': in_refmodel, 'd_in_reffree': in_reffree,
                        '~~>': '~~>',
                        'replacement': good_mrid, 'r_formula': good_formula, 'r_charge': good_charge, 'r_in_refmodel': good_in_refmodel, 'r_in_reffree': good_in_reffree}
                    results_df.append(new_row)


                    # populate the t-to-1 dictionary
                    to_translate_11[mrid] = good_mrid
                
                
                # for reactions
                elif mrmode == 'r': 
                    
                    # for this reaction to replace
                    r = model.reactions.get_by_id(mrid)
                    in_refmodel = mrid in mrids_refmodel
                    in_reffree = mrid in mrids_reffree
                    mids_involved = set([m.id for m in r.metabolites if m.id.rsplit('_', 1)[0] != 'h'])  # exclude protons.
                    gids_involved = set([g.id for g in r.genes])  
                    
                    # for the replacement reaction
                    good_r = model.reactions.get_by_id(good_mrid)
                    good_in_refmodel = good_mrid in mrids_refmodel
                    good_in_reffree = good_mrid in mrids_reffree
                    good_mids_involved = set([m.id for m in good_r.metabolites if m.id.rsplit('_', 1)[0] != 'h'])  # exclude protons.
                    good_gids_involved = set([g.id for g in good_r.genes])  
                    
                    
                    # determine if exaclty same reactants (except for protons).
                    # WARNING: this way we exclude transporters involved in different compartments
                    if mids_involved != good_mids_involved:
                        continue
                    
                    
                    # WARNING: by design, skip if both reactions are ancoded in the refmodel:
                    if in_refmodel and good_in_refmodel:
                        continue
                        
                        
                    # WARNING: by design, if two equivalment reactions have different gene sets, glue the gprs:
                    same_geneset = gids_involved == good_gids_involved
                        
                        
                    # log the future edits: 
                    new_row = {
                        'duplicated': mrid, 'd_reaction': r.reaction, 'd_gpr': r.gene_reaction_rule, 'd_in_refmodel': in_refmodel, 'd_in_reffree': in_reffree,
                        '~~>': '~~>',
                        'replacement': good_mrid, 'r_reaction': good_r.reaction, 'r_gpr': good_r.gene_reaction_rule, 'r_in_refmodel': good_in_refmodel, 'r_in_reffree': good_in_reffree, 'same_geneset': same_geneset}
                    results_df.append(new_row)


                    # populate the t-to-1 dictionary
                    to_translate_11[mrid] = good_mrid
                
     
    # produce the final results table: 
    results_df = pnd.DataFrame.from_records(results_df)
    
    
    # erease meaningless fileds when not using a reference
    if refmodel.id == '__EMPTY__': 
        results_df['d_in_refmodel'] = '-'
        results_df['r_in_refmodel'] = '-'
        
        
    # save tabular results
    if   mrmode=='m': results_df.to_csv('working/duplicates/dup_m_edits.csv')  
    elif mrmode=='r': results_df.to_csv('working/duplicates/dup_r_edits.csv')  
        
    return to_translate_11



def translate_targets(logger, model, to_translate, mancor):
    # attemps to rewrite reactions in a model, in a way that duplicate metabolites are solved.

    
    # define key objects: 
    modeled_mids  = [m.id for m in model.metabolites]
    results_df = []  # tabular results
    rs_to_remove = []  # to limit stoichiometric inconsistencies
    
    
    # for each reaction in the model, solve each eventual duplicate
    for r in model.reactions:
        
        # get mids, puremids, compartements, for each involved metabolite. 
        r_mids     = [m.id                   for m in r.metabolites]
        r_puremids = [m.id.rsplit('_', 1)[0] for m in r.metabolites]
        r_comps  =   [m.id.rsplit('_', 1)[1] for m in r.metabolites]
        
        
        # get number of metabolite IDs involved (compartment included). 
        # It will later used to check the presence of the exact same metabolite to both the sides of the reaction.
        # In these cases, the reaction will be removed to prevent stoichiometric inconsistency. 
        rids_involved_old = set([m.id for m in r.metabolites])
        
        
        # define 'old' and future reaction strings
        reaction_old = r.reaction 
        reaction_new = r.reaction
        reaction_new = ' '+reaction_new+' '  # for each mid to be surrounded by spaces.
        
        
        # record new metabolites defined in order to translate this reaction
        added_mids = []
        # duplicate ~~> translated mids in this reaction
        translated_mids = []
        
        
        # iterate thrugh each duplicate-replacement pair: 
        for duplicate in to_translate.keys(): 
            replacement = to_translate[duplicate]
            

            # check if this reaction contains the metabolite to change: 
            if duplicate in r_puremids: 
                translated_mids.append(f'{duplicate} ~~> {replacement}')
                

                # iterate involved metabolites until we reach the one to change: 
                for mid, puremid, comp in zip(r_mids, r_puremids, r_comps): 
                    if puremid == duplicate: 
                        # check if the new metabolite already exists.
                        if f'{replacement}_{comp}' not in modeled_mids:

                            # retrive the metabolite to use as template (coming from another compartment)
                            good_m = get_first_occurrence(replacement, model)
                            # create new metabolite:
                            new_m = cobra.Metabolite(f'{replacement}_{comp}')
                            model.add_metabolites([new_m])
                            # copy formula, charge and annotation
                            new_m = model.metabolites.get_by_id(f'{replacement}_{comp}')
                            new_m.formula = good_m.formula
                            new_m.charge  = good_m.charge
                            new_m.annotation = good_m.annotation
                            # set the right compartment
                            new_m.compartment = comp

                            # update the 'modeled_mids'
                            modeled_mids.append(f'{replacement}_{comp}')
                            
                            # record new metabolites needed by this translation
                            added_mids.append(f'{replacement}_{comp}')
                            

                        # edit the reaction
                        reaction_new = reaction_new.replace(' '+mid+' ', ' '+f'{replacement}_{comp}'+' ')
                        
                        
        # remove the spaces
        reaction_new = reaction_new[1:-1]
        
        if reaction_new != reaction_old:  # if some changes were made:
            
            
            # apply manual corrections if any 
            if r.id in mancor['reactions'].keys(): reaction_new = mancor['reactions'][r.id]
            
            
            # apply updated reaction
            r.build_reaction_from_string(reaction_new)
            
            
            # compute again the metabolites involved to prevent stoichiometric inconsistencies
            to_remove = '-'
            rids_involved_new = set([m.id for m in r.metabolites])
            if len(rids_involved_new) != len(rids_involved_old):  # number of metabolite IDs should not change. 
                rs_to_remove.append(r)
                to_remove = 'removed'


            # populate tabular results
            translated_mids =  f'{len(translated_mids)}: {translated_mids}'
            added_mids = f'{len(added_mids)}: {added_mids}'
            if added_mids == '0: []': added_mids = '-'
            bal_suggestions = str(r.check_mass_balance())
            if bal_suggestions == '{}': bal_suggestions = '-'
            if r.id.startswith('EX_') and len(r.metabolites)==1: bal_suggestions = '-'
            results_df.append({
                'rid': r.id, 'reaction_old': reaction_old, '~~>': '~~>', 'reaction_new': reaction_new, 
                'translated_mids': translated_mids, 'added_mids': added_mids, 'bal_suggestions': bal_suggestions, 'to_remove': to_remove})
    
    
    # remove reactions if needed
    model.remove_reactions(rs_to_remove)
    
    
    # save tabular results: 
    results_df = pnd.DataFrame.from_records(results_df)
    results_df.to_csv('working/duplicates/dup_m_translations.csv')



def parse_disconnected_metabolites(logger, model):
    
    
    # remove metabolites involved to 0 reactions. 
    to_remove = []
    for m in model.metabolites:
        if len(m.reactions) == 0: 
            to_remove.append(m)
    model.remove_metabolites(to_remove)
    logger.debug(f"Removed {len(to_remove)} disconnected metabolites: {[m.id for m in to_remove]}.")



def remove_duplicated_and_set_gpr(model, to_translate): 
    
    
    # first delete duplicated reactions: 
    to_delete = [model.reactions.get_by_id(rid) for rid in to_translate.keys()]
    model.remove_reactions(to_delete)
    
    
    # then update gpr where needed: 
    results_df = pnd.read_csv('working/duplicates/dup_r_edits.csv', index_col=0)
    results_df = results_df[results_df['same_geneset']==False]
    results_df = results_df.reset_index(drop=True)
    groups = results_df.groupby('replacement').groups
    for replacement in groups:
        group = results_df.iloc[groups[replacement], ]
        gprs = set(group['d_gpr'].to_list() + group['r_gpr'].to_list())
        final_gpr = '(' + ') or ('.join(list(gprs) ) + ')'
        r = model.reactions.get_by_id(replacement)
        r.gene_reaction_rule = final_gpr
        r.update_genes_from_gpr() 
                                                  


def solve_duplicates(logger, identity, coverage, refmodel, mancor_filepath):
    
    
    # log some message: 
    logger.info("Detecting duplicates using MetaNetX annotations...")
    
    
    # create subdirs without overwriting: 
    os.makedirs('working/duplicates/', exist_ok=True)
    
    
    # check presence of already computed files 
    if os.path.exists(f'working/duplicates/draft_panmodel.json') and os.path.exists(f'working/duplicates/md5.pickle'):
        if os.path.exists(f'working/duplicates/draft_panmodel_da.json') and os.path.exists(f'working/duplicates/md5_da.pickle'):
            if os.path.exists(f'working/duplicates/draft_panmodel_da_dd.json') and os.path.exists(f'working/duplicates/md5_da_dd.pickle'):
                with open('working/duplicates/md5.pickle', 'rb') as handler:
                    md5 = pickle.load(handler)
                with open('working/duplicates/md5_da.pickle', 'rb') as handler:
                    md5_da = pickle.load(handler)
                with open('working/duplicates/md5_da_dd.pickle', 'rb') as handler:
                    md5_da_dd = pickle.load(handler)
                # compare md5:
                if md5 == md5_da == md5_da_dd == get_md5_string('working/duplicates/draft_panmodel.json'):
                    # log some message: 
                    logger.info('Found all the needed files already computed. Skipping this step.')
                    # signal to skip this module:
                    return 0
    
    
    # load the needed models: 
    draft_panmodel = cobra.io.load_json_model('working/duplicates/draft_panmodel_da.json')
    reffree = cobra.io.load_json_model(f'working/free/draft_panmodel_{identity}_{coverage}.json')
    if refmodel != '-':
        refmodel_basename = os.path.basename(refmodel)
        refmodel = cobra.io.load_json_model(f'working/brh/{refmodel_basename}.refmodel_translated.json')
    else: refmodel = cobra.Model('__EMPTY__')
    
    
    # get the manual corrections dictionary:
    if mancor_filepath != '-': # existence of the file was verified above
        response = mancor_to_dict(logger, mancor_filepath)  # convert the filepath to dict
        if type(response) == int:
            if response == 1: return 1
        else: # good corrections file: 
            mancor = response  
            logger.info(  # log some message
                f"Using the provided manual corrections ({len(mancor['formulas'].keys())} formula corrections, {len(mancor['charges'].keys())} charge corrections, " + \
                f"{len(mancor['reactions'].keys())} reaction corrections, and {len(mancor['blacklist'])} reactions in blacklist)...")
    else: mancor = {'formulas': {}, 'charges': {}, 'reactions': {}, 'blacklist': []}  # emtpy, easier to handle
        
    
    # solve duplicated metabolites:
    logger.info("Detecting duplicate metabolites using MetaNetX annotations...")
    to_translate = get_translation_dictionary_mnx(draft_panmodel, mrmode='m')  # 1-to-many 
    to_translate = sort_translation_dictionary(to_translate, draft_panmodel, refmodel, reffree, mrmode='m')  # 1-to-1
    translate_targets(logger, draft_panmodel, to_translate, mancor)
    parse_disconnected_metabolites(logger, draft_panmodel)
    
    
    ### TMP
    from memote.support.consistency import find_unconserved_metabolites
    print(find_unconserved_metabolites(draft_panmodel))
    ### TMP
    
    
    
    # solve duplicated reactions:
    logger.info("Detecting duplicate reactions using MetaNetX annotations...")
    to_translate = get_translation_dictionary_mnx(draft_panmodel, mrmode='r')  # 1-to-many 
    to_translate = sort_translation_dictionary(to_translate, draft_panmodel, refmodel, reffree, mrmode='r')  # 1-to-1
    remove_duplicated_and_set_gpr(draft_panmodel, to_translate)
    
    
    # save deduplicated model
    cobra.io.save_json_model(draft_panmodel, f'working/duplicates/draft_panmodel_da_dd.json')
    
        
    # trace the parent model md5:
    parent_md5 = get_md5_string('working/duplicates/draft_panmodel.json')
    with open('working/duplicates/md5_da_dd.pickle', 'wb') as handle:
        pickle.dump(parent_md5, handle)
        
        
    return 0




def reset_growth_env(model):
    """Set all the EX_change reactions to (0, 1000).
    
    Args:
        model (cobra.Model): target model.

    """
    
    for r in model.reactions:
        if len(r.metabolites)==1 and r.id.startswith('EX_'):
            r.bounds = (0, 1000)
            
            
            
def set_bounded_uptakes(model, uptakes):
    """Set uptake of nutrients.
    
    Can be expressed as either concentrations [mmol/L] or fluxes [mmol/(h*gDW)], depending on your aims.
    
    Args:
        model (cobra.Model): target model.
        secretions (dict): dictionary keyed by EX_change reaction ID. Expressing values as tuples, the user can provide the standard deviation.

    """
    
    modeled_rids = [r.id for r in model.reactions]
    for ex_rid in uptakes.keys():
        if ex_rid not in modeled_rids: 
            raise Exception(f"{ex_rid} does not exists in this model.")
        
        
        # handle standard deviation
        if type(uptakes[ex_rid]) == tuple:
            model.reactions.get_by_id(ex_rid).lower_bound = -uptakes[ex_rid][0] -uptakes[ex_rid][1]
            model.reactions.get_by_id(ex_rid).upper_bound = -uptakes[ex_rid][0] +uptakes[ex_rid][1]
        else:  # no std was provided:
            model.reactions.get_by_id(ex_rid).lower_bound = -uptakes[ex_rid]
        
        
        
def set_bounded_secretions(model, secretions):
    """Set secretion of metabolites.
    
    Can be expressed as either concentrations [mmol/L] or fluxes [mmol/(h*gDW)], depending on your aims.
    
    Args:
        model (cobra.Model): target model.
        secretions (dict): dictionary keyed by EX_change reaction ID. Expressing values as tuples, the user can provide the standard deviation.

    """
    
    modeled_rids = [r.id for r in model.reactions]
    for ex_rid in secretions.keys():
        if ex_rid not in modeled_rids: 
            raise Exception(f"{ex_rid} does not exists in this model.")
            
            
        # handle standard deviation
        if type(secretions[ex_rid]) == tuple:
            model.reactions.get_by_id(ex_rid).lower_bound = secretions[ex_rid][0] -secretions[ex_rid][1]
            model.reactions.get_by_id(ex_rid).upper_bound = secretions[ex_rid][0] +secretions[ex_rid][1]
        else:  # no std was provided:
            model.reactions.get_by_id(ex_rid).lower_bound = secretions[ex_rid]
        
        
        
def set_unbounded_exchanges(model, exrs):
    """Define which melecules require a free exchange (-1000, 1000).
    
    Args:
        model (cobra.Model): target model.
        exrs (list): IDs of EX_change reactions to unbound.

    """
    
    modeled_rids = [r.id for r in model.reactions]
    for ex_rid in exrs:
        if ex_rid not in modeled_rids: 
            raise Exception(f"{ex_rid} does not exists in this model.")
        model.reactions.get_by_id(ex_rid).bounds = (-1000, 1000)


import pickle
import os
import multiprocessing
import itertools


import pandas as pnd
from Bio import SeqIO, SeqRecord
import cobra
import seaborn as sb
import matplotlib.pyplot as plt
from matplotlib.patches import Patch


from ..commons import chunkize_items
from ..commons import load_the_worker
from ..commons import gather_results
from ..commons import read_refmodel



def create_panmodel_proteome(logger, outdir):
    
    logger.info("Creating the reference proteome for the draft pan-model...")
    
    
    
    # A) from the panmodel
    
    # read the final draft panmodel
    draft_panmodel = read_refmodel(outdir + 'draft_panmodel.json')
    genes_to_report = set()
    for g in draft_panmodel.genes:
        if g.id == 'spontaneous':
            continue
        genes_to_report.add(g.id)
        
    
    # collect the reference sequences
    sr_list = []
    added = set()
    for record in SeqIO.parse('working/clustering/representatives.ren.faa', "fasta"):
        cluster, cds, accession = record.description.split(' ')
        if cluster in genes_to_report:
            sr = SeqRecord.SeqRecord(record.seq, id=cluster, description=f'{cds} {accession}')
            sr_list.append(sr)
            added.add(cluster)
            genes_to_report = genes_to_report - added
            
            
    # if all the sequences were recovered, write the fasta:
    if genes_to_report == set():
        with open(outdir + 'draft_panmodel.faa', 'w') as w_handler:
            count = SeqIO.write(sr_list, w_handler, "fasta")
        logger.debug(f"{len(added)} reference sequences written to " + outdir + 'draft_panmodel.faa' + '.')
        
        
        
    # B) from the PAM
    """
    # read the final draft panmodel
    pam = pnd.read_csv(outdir + 'pam.csv', index_col=0)
    genes_to_report = set()
    for cluster, row in pam.iterrows():
        genes_to_report.add(cluster)
        
    
    # collect the reference sequences
    sr_list = []
    added = set()
    for record in SeqIO.parse('working/clustering/representatives.ren.faa', "fasta"):
        cluster, cds, accession = record.description.split(' ')
        if cluster in genes_to_report:
            sr = SeqRecord.SeqRecord(record.seq, id=cluster, description=f'{cds} {accession}')
            sr_list.append(sr)
            added.add(cluster)
            genes_to_report = genes_to_report - added
            
            
    # if all the sequences were recovered, write the fasta:
    if genes_to_report == set():
        with open(outdir + 'draft_panproteome.faa', 'w') as w_handler:
            count = SeqIO.write(sr_list, w_handler, "fasta")
        logger.debug(f"{len(added)} reference sequences written to " + outdir + 'draft_panproteome.faa' + '.')
    """
            


def create_report(logger, outdir): 
    
    report = []  # list of dicts, future dataframe
    
    
    # get the retained genomes/proteomes (post filtering):
    with open('working/proteomes/species_to_proteome.pickle', 'rb') as handler:
        species_to_proteome = pickle.load(handler)
        for species in species_to_proteome.keys(): 
            for proteome in species_to_proteome[species]:
                basename = os.path.basename(proteome)
                accession, _ = os.path.splitext(basename)
                
                
                # populate the table: 
                report.append({'species': species, 'accession': accession})
                
                
    # save to file
    report = pnd.DataFrame.from_records(report)
    report.to_csv(outdir + 'report.csv')
    
    
    
    return 0 
    
    

def get_accession_to_recovery(logger, pam_modeled):
    
    # parse the pam (modeled clusters) to obtain metrics on gene recovery, per accession. 
    
    acc_to_recovery = {}
    for acc in pam_modeled.columns:
        acc_to_recovery[acc] = {
            'healthy': set(), 'recovered': set(),
            'frag': set(), 'refound': set(), 'overlap': set(), 
            'stop': set()}
        
        for index, row in pam_modeled.iterrows():
            cell = pam_modeled.loc[index, acc]
            if type(cell) == float: 
                continue

            one_healthy = False
            only_stop = True
            is_broken, is_refound, is_overlap = False, False, False

            for cds in cell.split(';'):
                if all(i not in cds for i in ['_stop', '_frag', '_refound', '_overlap']):
                    one_healthy = True
                if '_stop' not in cds: 
                    only_stop = False

                if '_frag' in cds: 
                    is_broken = True
                if '_refound' in cds:
                    is_refound = True
                if '_overlap' in cds: 
                    is_overlap = True

            if only_stop:    acc_to_recovery[acc]['stop'].add(index)
            else: 
                if one_healthy: acc_to_recovery[acc]['healthy'].add(index)
                else:
                    if is_broken:  acc_to_recovery[acc]['frag'].add(index)
                    if is_refound:  acc_to_recovery[acc]['refound'].add(index)
                    if is_overlap:  acc_to_recovery[acc]['overlap'].add(index)

                    if is_broken or is_refound or is_overlap:
                        acc_to_recovery[acc]['recovered'].add(index)
                        
        # could be parallelized!
        logger.debug(f"PAM parsed for {acc}.")

        
    return acc_to_recovery



def figure_genes_recovered(logger, outdir, pam_modeled):
    
    logger.info("Gathering metrics for gene recovery...")
    
    # get the dictionary of sets summarizing gene recovery 
    # (eg: acc_to_recovery['GCA_01'] = {'frag': {'Cluster_2', 'Cluster_3'}}
    acc_to_recovery = get_accession_to_recovery(logger, pam_modeled)
    
    
    # convert dictionary of sets to dataframe of numbers indexed by accession: 
    # plotting-dataframe generation: 
    rec_summary = [{'accession': acc} for acc, i in acc_to_recovery.items()]
    for i, acc in enumerate(acc_to_recovery.keys()): 
        for key, values in acc_to_recovery[acc].items(): rec_summary[i][key] = len(values)
    rec_summary = pnd.DataFrame.from_records(rec_summary)
    rec_summary = rec_summary.set_index('accession', drop=True, verify_integrity=True)
    
    
    logger.info("Producing figure for gene recovery in {outdir}/figures/genes_recovered.png...")
    
    
    # join the dataframe to get 'strain_isolate' and 'organism_name' fields.
    genomes_df = pnd.read_csv('working/genomes/genomes.csv', index_col=0)
    genomes_df = genomes_df.set_index('assembly_accession', drop=True, verify_integrity=True)
    # retain only quality-filtered genomes retaining the original order: 
    genomes_df = genomes_df.loc[[i for i in genomes_df.index if i in rec_summary.index.to_list()], ]   
    df = pnd.concat([genomes_df, rec_summary], axis=1)
                                
    
    # define colors:
    df = df.set_index('strain_isolate', drop=False)
    colors = df['organism_name'].map({species: f'C{number}' for number, species in enumerate(df['organism_name'].unique())}).to_dict()
        
    # draw bars:
    fig, ax = plt.subplots()
    _ = sb.barplot(df, x='strain_isolate', y='healthy', color='lightgrey', ax=ax)
    _ = sb.barplot(df, x='strain_isolate', y='recovered', color='grey', bottom=df['healthy'], ax=ax)
    
    # set tick labels
    ax.tick_params(axis='x', labelrotation=90)
    [label.set_color(colors[label.get_text()]) for label in ax.get_xticklabels()]
    
    # set legends:
    l1 = plt.legend(handles=[Patch(color=color, label=metric) for color, metric in zip(['grey','lightgrey'], ['Recovered','Healthy'])], title='', loc='upper left', bbox_to_anchor=(1.05, 0.5))
    l2 = plt.legend(handles=[Patch(color=f'C{number}', label=species) for number, species in enumerate(df['organism_name'].unique())], title='', loc='lower left', bbox_to_anchor=(1.05, 0.5))
    ax.add_artist(l1)  # l2 implicitly replaces l1
    
    ax.figure.set_size_inches(0.2*len(df), 4)
    ax.set_ylabel('modeled gene clusters')
    sb.despine()

    try: 
        plt.savefig(outdir + 'figures/genes_recovered.png', dpi=300, bbox_inches='tight')
    except:  # the png image could be too large, so we produce svg
        logger.info(f"PNG was too large: producing the SVG version instead {outdir}/figures/genes_recovered.svg...")
        plt.savefig(outdir + 'figures/genes_recovered.svg', bbox_inches='tight')


    
def get_species_to_core(pam_modeled, report):
    
    # get core genes of each species (dict of sets)
    species_to_core = {}
    
    for species in report['species'].unique(): 
        species_to_core[species] = set()  # row indexes
        
        accs = report[report['species']==species]['accession'].to_list()

        for index, row in pam_modeled[accs].iterrows():
            is_core = True
            for acc in accs:   # iterate over accessions for this species
                cell = row[acc]
                if type(cell) == float: 
                    is_core = False
                    break
                else:
                    only_stop = True
                    for cds in cell.split(';'):
                        if '_stop' not in cds:
                            only_stop = False
                    if only_stop:
                        is_core = False
                        break
            if is_core:
                species_to_core[species].add(index)
                
    return species_to_core
    
    

def get_overall_core(pam_modeled):
    
    # get core genes (considering all quality-filtered genomes in input)
    overall_core = set()
    
    for index, row in pam_modeled.iterrows():
        is_core = True
        for acc in pam_modeled.columns:   # iterate over accessions for this species
            cell = row[acc]
            if type(cell)==float: 
                is_core = False
            else:
                only_stop = True
                for cds in cell.split(';'):
                    if '_stop' not in cds:
                        only_stop = False
                if only_stop:
                    is_core = False
        if is_core:
            overall_core.add(index)
            
    return overall_core



def get_accession_to_genes(pam_modeled):

    # get all gene clusters of each quality-filtered genome (dict of sets): 
    strain_to_genes = {}
    
    for acc in pam_modeled.columns:
        strain_to_genes[acc] = set()
        for index, row in pam_modeled.iterrows(): 
            cell = pam_modeled.loc[index, acc]  
            
            if type(cell) != float:
                only_stop = True
                for cds in cell.split(';'):
                    if '_stop' not in cds:
                        only_stop = False
                if not only_stop:
                    strain_to_genes[acc].add(index)
                    
    return strain_to_genes
    
    
    
def figure_modeled_genes(logger, outdir, pam_modeled, report, draft_panmodel):
    
    logger.info("Gathering metrics for modeled gene clusters...")
    
    
    # get main assets:
    overall_core = get_overall_core(pam_modeled)
    species_to_core = get_species_to_core(pam_modeled, report)
    acc_to_genes = get_accession_to_genes(pam_modeled)
    

    # plotting-dataframe generation: 
    gene_summary = []   # list of dicts, future df
    for accession in acc_to_genes.keys():
        species = report.set_index('accession', drop=True, verify_integrity=True).loc[accession, 'species']
        gene_summary.append({
            'accession': accession,
            'accessory': len(acc_to_genes[accession] - overall_core.union(species_to_core[species])),
            'species_core': len(species_to_core[species] - overall_core),
            'overall_core': len(overall_core),
        })
    gene_summary = pnd.DataFrame.from_records(gene_summary).set_index('accession', drop=True, verify_integrity=True)
    
    
    logger.info("Producing figure for modeled gene clusters in {outdir}/figures/genes_modeled.png...")
    
    
    # join the dataframe to get 'strain_isolate' and 'organism_name' fields.
    genomes_df = pnd.read_csv('working/genomes/genomes.csv', index_col=0)
    genomes_df = genomes_df.set_index('assembly_accession', drop=True, verify_integrity=True)
    # retain only quality-filtered genomes retaining the original order: 
    genomes_df = genomes_df.loc[[i for i in genomes_df.index if i in gene_summary.index.to_list()], ]   
    df = pnd.concat([genomes_df, gene_summary], axis=1)
        
    
    # handle the draft pan-GSMM in a dedicated dataframe: 
    df_pan = df.copy()
    df_pan.loc['draft pan-GSMM'] = 0  # row
    df_pan['pan'] = 0   # column
    df_pan.loc['draft pan-GSMM', 'pan'] = len(draft_panmodel.genes)   # add information about the draft pan-GSMM in a dedicated column
    df_pan.loc['draft pan-GSMM', 'strain_isolate'] = 'draft pan-GSMM'
        
    
    # define colors:
    df = df.set_index('strain_isolate', drop=False)
    colors = df['organism_name'].map({species: f'C{number}' for number, species in enumerate(df['organism_name'].unique())}).to_dict()
    colors['draft pan-GSMM'] = 'black'
    
    # draw bars:
    fig, ax = plt.subplots()
    _ = sb.barplot(df, x='strain_isolate', y='overall_core', color='grey', ax=ax)
    _ = sb.barplot(df, x='strain_isolate', y='species_core', palette=colors, bottom=df['overall_core'], hue='strain_isolate', legend=False, ax=ax)
    _ = sb.barplot(df, x='strain_isolate', y='accessory', color='lightgrey', bottom=df['overall_core']+df['species_core'], ax=ax)
    _ = sb.barplot(df_pan, x='strain_isolate', y='pan', color='black', ax=ax)
    
    # set tick labels
    ax.tick_params(axis='x', labelrotation=90)
    [label.set_color(colors[label.get_text()]) for label in ax.get_xticklabels()]
    
    # set legends:
    l1 = plt.legend(handles=[Patch(color=color, label=metric) for color, metric in zip(['grey','lightgrey'], ['Common core','Non-core'])], title='', loc='upper left', bbox_to_anchor=(1.05, 0.5))
    handles = [Patch(color=f'C{number}', label=species) for number, species in enumerate(df['organism_name'].unique())]
    handles.append(Patch(color=f'black', label='draft pan-GSMM'))
    l2 = plt.legend(handles=handles, title='', loc='lower left', bbox_to_anchor=(1.05, 0.5))
    ax.add_artist(l1)  # l2 implicitly replaces l1
    
    ax.figure.set_size_inches(0.2*len(df), 4)
    ax.set_ylabel('modeled gene clusters')
    sb.despine()

    try: 
        plt.savefig(outdir + 'figures/genes_modeled.png', dpi=300, bbox_inches='tight')
    except:  # the png image could be too large, so we produce svg
        logger.info(f"PNG was too large: producing the SVG version instead {outdir}/figures/genes_modeled.svg...")
        plt.savefig(outdir + 'figures/genes_modeled.svg', bbox_inches='tight')
        
    

def task_rpam_parse_column(accession, args): 
        
    # retrive the arguments:
    draft_panmodel = args['draft_panmodel']
    pam_modeled = args['pam_modeled']


    # create a copy of the model
    ss_model = draft_panmodel.copy()

    # define which genes to remove: 
    to_remove = []
    for cluster, row in pam_modeled.iterrows():
        cell = pam_modeled.loc[cluster, accession]
        if type(cell)==float: 
            to_remove.append(ss_model.genes.get_by_id(cluster))
            continue
        only_stop = True
        for cds in cell.split(';'):
            if '_stop' not in cds:
                only_stop = False
        if only_stop:
            to_remove.append(ss_model.genes.get_by_id(cluster))
            continue
    cobra.manipulation.delete.remove_genes(ss_model, to_remove, remove_reactions=True)


    # create new row:
    row_dict = {'accession': accession}
    for r in ss_model.reactions: 
        row_dict[r.id] = 1


    # return the new dataframe row
    return [row_dict]
    
    
    
def get_rpam_from_panmodel(logger, cores, draft_panmodel, pam_modeled):
    # create a reaction presence/absence matrix starting from a PAM and a draft panmodel: 
    
    
    # create items for parallelization: 
    items = []
    for accession in pam_modeled.columns: 
        items.append(accession)
        
        
    # randomize and divide in chunks: 
    chunks = chunkize_items(items, cores)
    
    
    # initialize the globalpool:
    globalpool = multiprocessing.Pool(processes=cores, maxtasksperchild=1)
    
    
    # start the multiprocessing: 
    results = globalpool.imap(
        load_the_worker, 
        zip(chunks, 
            range(cores), 
            itertools.repeat(['accession'] + [r.id for r in draft_panmodel.reactions]), 
            itertools.repeat('accession'), 
            itertools.repeat(logger), 
            itertools.repeat(task_rpam_parse_column),
            itertools.repeat({'draft_panmodel': draft_panmodel, 'pam_modeled': pam_modeled}),
        ), chunksize = 1)
    all_df_combined = gather_results(results)  # all_df_combined can be ignored.
    
    
    # empty the globalpool
    globalpool.close() # prevent the addition of new tasks.
    globalpool.join()  
    
    
    # format the rpam
    rpam = all_df_combined
    # now replace missing values with 0. 
    # In 'older' pandas versions, fillna() was doing silent dowcasting (see https://medium.com/@felipecaballero/deciphering-the-cryptic-futurewarning-for-fillna-in-pandas-2-01deb4e411a1)
    # In 'future' pandas versions, silent downcasting will be forbidden. 
    # In 'middle' pandas versions, there is an option to simulate the future behaviour.
    # What follows should guarantee the old behaviour in 'older' and 'middle' pandas versions:
    try:
        with pnd.option_context('future.no_silent_downcasting', True):
            rpam = rpam.fillna(0).infer_objects()   # infer_objects() will do downcasting like in 'older' pandas versions
    except: # OptionError: "No such keys(s): 'future.no_silent_downcasting'" ==> 'older' pandas version
        rpam = rpam.fillna(0)  
    rpam = rpam.astype(int)  # force from float to int.
    rpam = rpam.T  # transpose: reactions as rows.
        
        
    return rpam
    


def get_species_to_core_rpam(rpam, report):
    # get core genes of each species (dict of sets)
    
    species_to_core = {}
    for species in report['species'].unique(): 
        species_to_core[species] = set()  # row indexes
        
        accs = report[report['species']==species]['accession'].to_list()

        for index, row in rpam[accs].iterrows():
            is_core = True
            for acc in accs:   # iterate over accessions for this species
                cell = row[acc]
                if cell == 0: 
                    is_core = False
                    break
            if is_core:
                species_to_core[species].add(index)
                
    return species_to_core


    
def get_overall_core_rpam(rpam):
    # given a rpam, get the core reactions (set): 
    
    
    overall_core = set()
    for index, row in rpam.iterrows():
        is_core = True
        for acc in rpam.columns:  
            cell = row[acc]
            if cell == 0: 
                is_core = False
        if is_core: 
            overall_core.add(index)
            
    return overall_core
    
    

def get_accession_to_reactions(rpam):
    # get reactions fore each quality-filtered input genome: 
    
    acc_to_reactions = {}
    for acc in rpam.columns:
        acc_to_reactions[acc] = set()
        for index, row in rpam.iterrows(): 
            cell = rpam.loc[index, acc]   # get columns (accs) for this species.
            if cell != 0:
                acc_to_reactions[acc].add(index)
                    
    return acc_to_reactions
    
    
    
def figure_modeled_reactions(logger, outdir, cores, pam_modeled, report, draft_panmodel):
    
    
    # get main assets: 
    rpam = get_rpam_from_panmodel(logger, cores, draft_panmodel, pam_modeled)
    overall_core_rpam = get_overall_core_rpam(rpam)
    species_to_core_rpam = get_species_to_core_rpam(rpam, report)
    acc_to_reactions = get_accession_to_reactions(rpam)

    
    # plotting-dataframe generation: 
    reaction_summary = []   # list of dicts, future df
    for accession in acc_to_reactions.keys():
        species = report.set_index('accession', drop=True, verify_integrity=True).loc[accession, 'species']
        reaction_summary.append({
            'accession': accession,
            'accessory': len(acc_to_reactions[accession] - overall_core_rpam.union(species_to_core_rpam[species])),
            'species_core': len(species_to_core_rpam[species] - overall_core_rpam),
            'overall_core': len(overall_core_rpam),
        })
    reaction_summary = pnd.DataFrame.from_records(reaction_summary).set_index('accession', drop=True, verify_integrity=True)
    
    
    logger.info("Producing figure for modeled reactions in {outdir}/figures/reactions_modeled.png...")
    
    
    # join the dataframe to get 'strain_isolate' and 'organism_name' fields.
    genomes_df = pnd.read_csv('working/genomes/genomes.csv', index_col=0)
    genomes_df = genomes_df.set_index('assembly_accession', drop=True, verify_integrity=True)
    # retain only quality-filtered genomes retaining the original order: 
    genomes_df = genomes_df.loc[[i for i in genomes_df.index if i in reaction_summary.index.to_list()], ]   
    df = pnd.concat([genomes_df, reaction_summary], axis=1)
        
    
    # handle the draft pan-GSMM in a dedicated dataframe: 
    df_pan = df.copy()
    df_pan.loc['draft pan-GSMM'] = 0  # row
    df_pan['pan'] = 0   # column
    df_pan.loc['draft pan-GSMM', 'pan'] = len(draft_panmodel.reactions)   # add information about the draft pan-GSMM in a dedicated column
    df_pan.loc['draft pan-GSMM', 'strain_isolate'] = 'draft pan-GSMM'
        
    
    # define colors:
    df = df.set_index('strain_isolate', drop=False)
    colors = df['organism_name'].map({species: f'C{number}' for number, species in enumerate(df['organism_name'].unique())}).to_dict()
    colors['draft pan-GSMM'] = 'black'
    
    # draw bars:
    fig, ax = plt.subplots()
    _ = sb.barplot(df, x='strain_isolate', y='overall_core', color='grey', ax=ax)
    _ = sb.barplot(df, x='strain_isolate', y='species_core', palette=colors, bottom=df['overall_core'], hue='strain_isolate', legend=False, ax=ax)
    _ = sb.barplot(df, x='strain_isolate', y='accessory', color='lightgrey', bottom=df['overall_core']+df['species_core'], ax=ax)
    _ = sb.barplot(df_pan, x='strain_isolate', y='pan', color='black', ax=ax)
    
    # set tick labels
    ax.tick_params(axis='x', labelrotation=90)
    [label.set_color(colors[label.get_text()]) for label in ax.get_xticklabels()]
    
    # set legends:
    l1 = plt.legend(handles=[Patch(color=color, label=metric) for color, metric in zip(['grey','lightgrey'], ['Common core','Non-core'])], title='', loc='upper left', bbox_to_anchor=(1.05, 0.5))
    handles = [Patch(color=f'C{number}', label=species) for number, species in enumerate(df['organism_name'].unique())]
    handles.append(Patch(color=f'black', label='draft pan-GSMM'))
    l2 = plt.legend(handles=handles, title='', loc='lower left', bbox_to_anchor=(1.05, 0.5))
    ax.add_artist(l1)  # l2 implicitly replaces l1
    
    ax.figure.set_size_inches(0.2*len(df), 4)
    ax.set_ylabel('modeled reactions')
    sb.despine()

    try: 
        plt.savefig(outdir + 'figures/reactions_modeled.png', dpi=300, bbox_inches='tight')
    except:  # the png image could be too large, so we produce svg
        logger.info(f"PNG was too large: producing the SVG version instead {outdir}/figures/reactions_modeled.svg...")
        plt.savefig(outdir + 'figures/reactions_modeled.svg', bbox_inches='tight')
            

    
def create_recon_plots(logger, outdir, cores, nofig):
    
    
    logger.info("Producing figures for preliminary reconstruction metrics...")
    
    # load main assets:
    draft_panmodel = read_refmodel(outdir + 'draft_panmodel.json')
    pam = pnd.read_csv(outdir + 'pam.csv', index_col=0)
    report = pnd.read_csv(outdir + 'report.csv', index_col=0)
    
    
    # filter pam for clusters modeled in draft_panmodel
    modeled_gids = [g.id for g in draft_panmodel.genes if g.id.startswith('Cluster_')]
    pam_modeled = pam.loc[modeled_gids, ]
    
    
    # make 3 plots:
    if not nofig: 
        figure_genes_recovered(logger, outdir, pam_modeled)
        figure_modeled_genes(logger, outdir, pam_modeled, report, draft_panmodel)
        figure_modeled_reactions(logger, outdir, cores, pam_modeled, report, draft_panmodel)
    
    
    return 0
    
    


import random
from importlib import resources
import os
import pickle
import glob
import subprocess
import hashlib
import json


import cobra
import pandas as pnd
from Bio import SeqIO, SeqRecord, Seq


from gempipe.curate.medium import reset_growth_env




def chunkize_items(items, cores):
    
    
    # divide items in chunks: 
    random.shuffle(items)  # randomly re-order items
    nitems_inchunk = int(len(items) / cores)
    if len(items) % cores !=0: nitems_inchunk += 1
    chunks = [items[x *nitems_inchunk : (x+1)* nitems_inchunk] for x in range(cores)]
    
    
    return chunks



def load_the_worker(arguments):
        
        
    # get the arguments:
    items = arguments[0]
    worker = arguments[1] +1   
    columns = arguments[2]
    index = arguments[3]
    logger = arguments[4]
    function = arguments[5]
    args = arguments[6]


    # iterate over each item: 
    df_combined = []
    cnt_items_processed = 0
    for item in items:


        # perform the annotation for this genome: 
        new_rows = function(item, args)
        df_combined = df_combined + new_rows
        

        # notify the logging process: 
        cnt_items_processed += 1
        logger.debug(f"W#{worker}-PID {os.getpid()}: {round(cnt_items_processed/len(items)*100, 1)}%")


    # join the tabular results of each item:
    if df_combined == []:  # this worker was started empty.
        df_combined = pnd.DataFrame(columns = columns)
    else: df_combined = pnd.DataFrame.from_records(df_combined)
    df_combined = df_combined.set_index(index, verify_integrity=True)
    return df_combined



def gather_results(results):
    
    
    # perform final concatenation of the tabular results:
    all_df_combined = []
    for result in results: 
        if isinstance(result, pnd.DataFrame):
            all_df_combined.append(result)
    
    # handle the following pandas future warning:
    # """FutureWarning: The behavior of DataFrame concatenation with empty or all-NA entries is deprecated. \
    # In a future version, this will no longer exclude empty or all-NA columns when determining the result dtypes. \
    # To retain the old behavior, exclude the relevant entries before the concat operation."""
    # To handle this: (1) Filter out empty DataFrames. (2) Filter out DataFrames that contain only NA values.
    all_df_combined = [df for df in all_df_combined if not df.empty and not df.isna().all().all()]

    all_df_combined = pnd.concat(all_df_combined, axis=0)
    
    
    return all_df_combined



def get_retained_accessions():
    # to be called after the genomes filtering.
    
    
    accessions = set()
    with open('working/proteomes/species_to_proteome.pickle', 'rb') as handler:
        species_to_proteome = pickle.load(handler)
        for species in species_to_proteome.keys(): 
            for proteome in species_to_proteome[species]:
                basename = os.path.basename(proteome)
                accession, _ = os.path.splitext(basename)
                accessions.add(accession)
    return accessions



def check_cached(logger, pam_path, imp_files, summary_path=None):
    
    
    # get the accessions retained:
    accessions = get_retained_accessions()
    
    
    # search for the PAM: 
    if os.path.exists(pam_path):
        pam = pnd.read_csv(pam_path, index_col=0)
        columns = set(list(pam.columns))
        
        
        # search for the optional summary:
        if summary_path != None:
            if os.path.exists(summary_path): 
                summary = pnd.read_csv(summary_path, index_col=0)
                rows = set(list(summary.index))
            else: return None
        else: rows = columns
            
            
        # check if accessions are the same (no less, no more):
        if accessions == columns == rows:
            
            
            # check the presence of important files:
            if all([os.path.exists(i) for i in imp_files]):
                # log some message: 
                logger.info('Found all the needed files already computed. Skipping this step.')
                
                
                # signal to skip this module:
                return 0
                
            
    return None



def create_summary(logger, module_dir):
    
    
    # get the accessions retained:
    accessions = get_retained_accessions()
    
    
    # parse each results file: 
    summary = []
    for file in glob.glob(f'{module_dir}/results/*.csv'):
        accession = file.rsplit('/', 1)[1].replace('.csv', '')
        if accession not in accessions: 
            continue  # other files present from previous runs.
        with open(file, 'r') as r_handler: 
            if r_handler.read() == '""\n':  # if the result csv for this accession is empty: 
                summary.append({'accession': accession, 'n_refound': 0, 'n_frag': 0, 'n_overlap': 0, 'n_stop': 0})
                continue
                
                
        # populate the summary: 
        result = pnd.read_csv(file, sep=',', index_col=0)
        summary.append({
            'accession': accession, 
            'n_refound': len(result[result['ID'].str.contains('_refound')]), 
            'n_frag': len(result[result['ID'].str.contains('_frag')]), 
            'n_overlap': len(result[result['ID'].str.contains('_overlap')]), 
            'n_stop': len(result[result['ID'].str.contains('_stop')]),
        })

        
    # write the summary to disk
    summary = pnd.DataFrame.from_records(summary)
    summary = summary.set_index('accession', drop=True, verify_integrity=True)
    summary.to_csv(f'{module_dir}/summary.csv')
    
    
    return 0



def update_pam(logger, module_dir, pam):
    
    
    # get the accessions retained:
    accessions = get_retained_accessions()
    
    
    # define important objects:
    cnt_newgenes = 0
    pam_update = pam.copy()
    
    
    # parse each results file: 
    for file in glob.glob(f'{module_dir}/results/*.csv'):
        accession = file.rsplit('/', 1)[1].replace('.csv', '')
        if accession not in accessions: 
            continue  # other files present from previous runs.
        with open(file, 'r') as r_handler: 
            if r_handler.read() == '""\n':  # if the result csv for this accession is empty: 
                continue
        # populate the summary: 
        result = pnd.read_csv(file, sep=',', index_col=0)


        # update the PAM: 
        for cluster in set(result['cluster'].to_list()): 
            new_genes = result[result['cluster']==cluster]['ID'].to_list()
            cnt_newgenes += len(new_genes)
            cell = ';'.join(new_genes)
            pam_update.loc[cluster, accession] = cell
            
        
    # write the updated PAM to disk
    pam_update.to_csv(f'{module_dir}/pam.csv')
    
    
    # write some log messages:
    logger.debug(f'Added {cnt_newgenes} new sequences.')
    
    
    return 0



def extract_aa_seq_from_genome(db, contig, strand, start, end ):
    
    
    # execute the command
    command = f'''blastdbcmd -db {db} -entry "{contig}" -range "{start}-{end}" -outfmt "%s"'''
    process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
    process.wait()
    
    
    # read the output from stdout
    curr_stream = process.stdout.read()
    curr_stream = curr_stream.decode('utf8')  # See https://stackoverflow.com/q/41918836
    curr_stream = curr_stream.rstrip()  # remove end of line
    
    
    # reverse complement if on the negative strand
    seq = Seq.Seq(curr_stream)
    if strand == '-': 
        seq = seq.reverse_complement()

        
    # trim the sequences to make it multiple of three, otherwise I get the following warning: 
    # BiopythonWarning: Partial codon, len(sequence) not a multiple of three. 
    seq_trimmed = seq[:len(seq) - (len(seq) % 3)]


    # translate to stop:
    seq_translated = seq_trimmed.translate()
    seq_translated_tostop = seq_trimmed.translate(to_stop=True)
    
    
    return seq_translated, seq_translated_tostop



def get_blast_header():
    
    
    # to standardize all the blast subprocesses
    return "qseqid sseqid pident ppos length qlen slen qstart qend sstart send evalue bitscore qcovhsp scovhsp"



def get_md5_string(filepath):
    
    
    with open(filepath, 'rb') as file:  # 'rb' good also for txt files.
        md5 = hashlib.md5()
        md5.update(file.read())  # warning: no chunks; keep attention with large files.
        md5_string = md5.hexdigest()
    return md5_string



def read_refmodel(refmodel): 
    
    
    # filter to retain just modeled genes:
    if refmodel.endswith('.json'): 
        refmodel = cobra.io.load_json_model(refmodel)
    elif refmodel.endswith('.sbml'): 
        refmodel = cobra.io.read_sbml_model(refmodel)
    elif refmodel.endswith('.xml'): 
        refmodel = cobra.io.read_sbml_model(refmodel)
    else:
        logger.error(refmodel + ": extension not recognized.")
        return 1 
    return refmodel



def get_outdir(outdir):
    
    # create the main output directory: 
    if outdir.endswith('/') == False: outdir = outdir + '/'
    os.makedirs(outdir, exist_ok=True)
    
    return outdir



def get_media_definitions(logger, media_filepath):
    
    
    logger.debug("Loading the provided media definitions...")
    
    
    media_files = []  # filepaths to json media definitions
    # check if the user specified something:
    if media_filepath != '-':  
        # get the media files:
        if os.path.exists(media_filepath):
            # check if the user specified a folder:
            if os.path.isdir(media_filepath):
                if media_filepath[-1] != '/': media_filepath = media_filepath + '/'
                media_files = glob.glob(media_filepath + '*')
            else:  # the user specified a single file:
                media_files = [media_filepath]
        else:  # path does not exixts (no folder / no file)
            logger.error(f"The path provided for media definitions (-m/--media) does not exists: {media_filepath}.")
            return 1
    else:  # use a predefined built-in minimal medium
        logger.debug("No definitions provided: loading the built-in minimal aerobic medium...")
        with resources.path("gempipe.assets", "minimal_medium.json") as asset_path:  # taken from CarveMe v1.5.2
            media_files = [asset_path]
         
        
    # check the formatting of the media files, producing a single dict contianing all provided media:
    media = {}   # single dict for all media provided
    for file_path in media_files:
        try:
            with open(file_path, 'r') as file:
                medium_data = json.load(file)
                media[medium_data['name']] = {}
                for rid in medium_data['exchanges'].keys():
                    media[medium_data['name']][rid] = medium_data['exchanges'][rid]
        except json.JSONDecodeError as e:
            logger.error(f"The provided medium file (JSON format) {file_path} encountered the following decoding error: {e}.")
            return 1
        
        
    logger.debug(f"Loaded {len(media.keys())} media definitions: {', '.join(list(media.keys()))}.")
    return media
            
    
          
def apply_json_medium(model, medium):
    
    
    reset_growth_env(model)
    for rid in medium.keys():
        try: model.reactions.get_by_id(rid).lower_bound = medium[rid]
        except:  # rid not found inside the 'expanded_universe'
            return rid
        
    return 0



def check_panmodel_growth(logger, panmodel, media, minpanflux): 
    
    
    # log some message:
    logger.debug("Testing the growth of the draft panmodel on the provided media...")
    
    
    # iterate the provided media
    for medium_name, medium in media.items():


        # apply the medium recipe:
        response = apply_json_medium(panmodel, medium)
        if type(response)==str: 
            logger.error(f"The exchange reaction '{response}' contained in the medium definition '{medium_name}' does not exist in the expanded_universe.")
            return 1


        # verify the growth of the expanded_universe:
        res = panmodel.optimize()
        obj_value = res.objective_value
        status = res.status
        can_growth = res.status=='optimal' and obj_value >= minpanflux
        logger.debug(f"'panmodel' growth on {medium_name}: {can_growth} ({status}, {obj_value}).")


        # raise error if it cannot grow:
        if not can_growth:
            return False
        
        
    return True



def strenghten_uptakes(model): 
    
    
    exr_ori = {}
    for r in model.reactions: 
        # get just the exchange reactions: 
        if len(r.metabolites)==1 and list(r.metabolites)[0].id.rsplit('_', 1)[-1] == 'e': 
            if r.lower_bound < 0:
                exr_ori[r.id] = r.lower_bound
                r.lower_bound = -1000

    return exr_ori



def get_allmeta_df(): 
    
    
    genomes_df = pnd.read_csv('working/genomes/genomes.csv', index_col=0)
    genomes_df = genomes_df.set_index('assembly_accession', drop=True, verify_integrity=True)

    bmetrics_df = pnd.read_csv('working/filtering/bmetrics.csv', index_col=0)
    bmetrics_df = bmetrics_df.set_index('accession', drop=True, verify_integrity=True)

    tmetrics_df = pnd.read_csv('working/filtering/tmetrics.csv', index_col=0)
    tmetrics_df = tmetrics_df.set_index('accession', drop=True, verify_integrity=True)
    
    allmeta_df = pnd.concat([genomes_df, bmetrics_df, tmetrics_df], axis=1)
    
    allmeta_df = allmeta_df[['organism_name', 'strain_isolate', 'C', 'F', 'M', 'ncontigs', 'sum_len', 'N50']]
    return allmeta_df



import pickle
import os
import subprocess


import pandas as pnd
from Bio import SeqIO, SeqRecord, Seq


from ..commons import check_cached



def create_combined(logger):
    
    
    # read which proteomes passed the filters: 
    with open('working/proteomes/species_to_proteome.pickle', 'rb') as handler:
        species_to_proteome = pickle.load(handler)
    proteomes = []  # get the proteomes to parse
    for species in species_to_proteome.keys(): 
        for proteome in species_to_proteome[species]:
            proteomes.append(proteome)
        
           
    # create foundamental objects:
    sequences_df = []  # future proteins dataframe
    acc_to_seqs = {}  # accession-to-proteins dictionary (1-to-many)
    seq_to_acc = {}  # protein-to-accession dictionary (1-to-1)
             
    
    # parse each proteome:
    sr_list = []
    for proteome in proteomes:
        basename = os.path.basename(proteome)
        accession, _ = os.path.splitext(basename)


        # open the proteome and read its CDSs
        cnt_prots = 0
        with open(proteome, 'r') as cds_handler:
            for seqrecord in SeqIO.parse(cds_handler, "fasta"):
                seqid = seqrecord.id
                seq = seqrecord.seq
                sr = SeqRecord.SeqRecord(seq, id=seqid, description=accession)
                sr_list.append(sr)
                sequences_df.append({'accession': accession, 'cds': seqid, 'aaseq': str(seq)})
                cnt_prots += 1


                # populate the accession-to-proteins dict:
                if accession not in acc_to_seqs.keys(): 
                    acc_to_seqs[accession] = set()
                acc_to_seqs[accession].add(seqid)


                # populate the protein-to-accession dict:
                if seqid not in seq_to_acc.keys(): 
                    seq_to_acc[seqid] = accession
                else: 
                    logger.error(f"The 'seqid' should be unique ('seq_to_acc[{seqid}] = {accession}').")
                    return 1


            # some log messages:
            logger.debug(f"{accession}: {cnt_prots} initial coding sequences found.")


    # write a single fasta with all proteins from all strains
    os.makedirs('working/clustering/', exist_ok=True)
    with open('working/clustering/combined.faa', 'w') as w_handler: 
        count = SeqIO.write(sr_list, w_handler, "fasta")

        
        # write the protein dataframe to file: 
        sequences_df = pnd.DataFrame.from_records(sequences_df)
        sequences_df = sequences_df.set_index('cds', verify_integrity=True)
        sequences_df.to_csv('working/clustering/sequences.csv')
        
        
        # write dictionaries to file:
        with open('working/clustering/acc_to_seqs.pickle', 'wb') as file:
            pickle.dump(acc_to_seqs, file)
        with open('working/clustering/seq_to_acc.pickle', 'wb') as file:
            pickle.dump(seq_to_acc, file)
        
        
        # some log messages:
        logger.debug(f"Total initial coding sequences found: {len(sequences_df)}.")
        
        
        return 0

    

def perform_clustering(logger, cores): 
    
    
    # some log messages: 
    logger.debug("Clustering with " + str(cores) + " cores...")


    # launch the command: 
    with open(f'working/logs/stdout_clustering.txt', 'w') as stdout, open(f'working/logs/stderr_clustering.txt', 'w') as stderr:
        # -d 0: keep the entire sequence ID in the .clstr file
        command = f"""cd-hit \
            -T {cores} -M 0 \
            -i working/clustering/combined.faa \
            -o working/clustering/representatives.faa \
            -g 1 -aL 0.70 -aS 0.70 -c 0.90 -d 0"""
        process = subprocess.Popen(command, shell=True, stdout=stdout, stderr=stderr)
        process.wait()
        
    
    # some log messages: 
    logger.debug("Clustering completed.")
    logger.debug("Parsing the clusters...")
    
    
    # create foundamental objects:
    cluster_to_seqs = {}  # cluster to sequences (1-to-many)
    cluster_to_rep = {}  # cluster to representative (1-to-1)
    seq_to_cluster = {}  # sequence to cluster (1-to-1)
    
    
    # parse the clstr_file
    with open('working/clustering/representatives.faa.clstr', 'r') as handler:
        rows = handler.readlines()
        for row in rows: 
            
            
            # the row has a cluster name:
            if row.startswith('>'): 
                cluster = row[1:-1] # '-1' remove the '\n'
                cluster = cluster.replace(' ', '_') # 'Cluster 548' -> 'Cluster_548'
                if cluster not in cluster_to_seqs.keys(): 
                    cluster_to_seqs[cluster] = set()
                    
            
            # the row has a sequence: 
            else: 
                seq, tail = row.split('>',1)[1].rsplit('... ', 1)
                if tail == '*\n':  # if representative:
                    cluster_to_rep[cluster] = seq
                cluster_to_seqs[cluster].add(seq)
                seq_to_cluster[seq] = cluster
                

    # Write the dictionaries to file:
    with open('working/clustering/cluster_to_seqs.pickle', 'wb') as handler:
        pickle.dump(cluster_to_seqs, handler)
    with open('working/clustering/cluster_to_rep.pickle', 'wb') as handler:
        pickle.dump(cluster_to_rep, handler)
    with open('working/clustering/seq_to_cluster.pickle', 'wb') as handler:
        pickle.dump(seq_to_cluster, handler)
        
        
    # load previously created seq_to_acc dictionary: 
    with open('working/clustering/seq_to_acc.pickle', 'rb') as handler:
        seq_to_acc = pickle.load(handler)
        
        
    # create the dictionary rep-to-aaseq:
    rep_to_aaseq = {}
    with open('working/clustering/representatives.faa', 'r') as r_handler: 
        for seqrecord in SeqIO.parse(r_handler, "fasta"):
            seq = seqrecord.seq # <class 'Bio.Seq.Seq'>
            seqid = seqrecord.id
            rep_to_aaseq[seqid] = str(seq)
    with open('working/clustering/rep_to_aaseq.pickle', 'wb') as handler:
        pickle.dump(rep_to_aaseq, handler)
        
        
    # rename the representative sequences:
    with open('working/clustering/representatives.ren.faa', 'w') as w_handler:
        sr_list = []
        with open('working/clustering/representatives.faa', 'r') as r_handler: 
            for seqrecord in SeqIO.parse(r_handler, "fasta"):
                seq = seqrecord.seq # <class 'Bio.Seq.Seq'>
                seqid = seqrecord.id
                new_id = seq_to_cluster[seqid]
                accession = seq_to_acc[seqid]
                sr = SeqRecord.SeqRecord(seq, id=new_id, description=f'{seqid} {accession}')
                sr_list.append(sr)
        count = SeqIO.write(sr_list, w_handler, "fasta")
        
        
    # some log messages: 
    logger.debug("Parsing completed.")
    
    
    return 0



def create_pam(logger):
    
    
    # load previously created dictionaries: 
    with open('working/clustering/cluster_to_seqs.pickle', 'rb') as handler:
        cluster_to_seqs = pickle.load(handler)
    with open('working/clustering/seq_to_acc.pickle', 'rb') as handler:
        seq_to_acc = pickle.load(handler)
    
    
    # create the matrix
    rows = []
    for cluster in cluster_to_seqs.keys():
        
        
        # for each accession, determine the CDSs belonging to this Cluster.
        acc_to_seqs_found = {}
        for seq in cluster_to_seqs[cluster]: 
            accession = seq_to_acc[seq]
            if accession not in acc_to_seqs_found.keys():
                acc_to_seqs_found[accession] = set()
            acc_to_seqs_found[accession].add(seq)
            
        
        # append a row to the dataframe: 
        row = {'cluster': cluster}
        for accession in acc_to_seqs_found.keys():
            row[accession] = ';'.join(acc_to_seqs_found[accession])
        rows.append(row)
        
        
    # write the pam to disk
    pam = pnd.DataFrame.from_records(rows)
    pam = pam.set_index('cluster', verify_integrity=True)
    pam.to_csv('working/clustering/pam.csv')
    
    
    # some log messages: 
    logger.debug("Created an initial presence/absence matrix (PAM) with shape: " + str(pam.shape))
    
    
    # create accession to suffixes dictionary: 
    logger.debug("Storing the suffix for each genome...")
    acc_to_suffix = {}
    for acc in pam.columns:
        suffixes = [i for i in pam[acc].to_list() if type(i) != float]
        suffix = list(set([i.split('_', 1)[0] for i in suffixes]))[0]
        acc_to_suffix[acc] = suffix
    with open('working/clustering/acc_to_suffix.pickle', 'wb') as handler:
        pickle.dump(acc_to_suffix, handler)
        
        
    # create the cluster to relative frequency dictionary: 
    # with the following binary expression, eventual '_stop' are included.
    cluster_to_absfreq = pam.apply(lambda col: col.map(lambda x: 1 if (type(x) != float and x != '') else 0)).sum(axis=1)
    cluster_to_relfreq = round(cluster_to_absfreq / len(pam.columns) * 100, 1)
    cluster_to_relfreq = cluster_to_relfreq.to_dict()
    with open('working/clustering/cluster_to_relfreq.pickle', 'wb') as handler:
        pickle.dump(cluster_to_relfreq, handler)
    
    
    return 0
        


def compute_clusters(logger, cores): 
    
    
    # some log messages:
    logger.info("Clustering the proteins...")
    
    
    # check if it's everything pre-computed
    response = check_cached(
        logger, pam_path='working/clustering/pam.csv', 
        imp_files=[
            'working/clustering/acc_to_seqs.pickle',
            'working/clustering/cluster_to_rep.pickle',
            'working/clustering/cluster_to_seqs.pickle',
            'working/clustering/seq_to_acc.pickle',
            'working/clustering/seq_to_cluster.pickle',
            'working/clustering/acc_to_suffix.pickle',
            'working/clustering/cluster_to_relfreq.pickle',
            'working/clustering/rep_to_aaseq.pickle',
            'working/clustering/representatives.ren.faa',
            'working/clustering/sequences.csv'])
    if response == 0: return 0
    
    
    # combine together all the sequences from all the strains: 
    response = create_combined(logger)
    if response == 1: return 1
    
    
    # perform the clustering 
    response = perform_clustering(logger, cores)
    if response == 1: return 1
    
    
    # creating the presence/absence matrix (PAM)
    response = create_pam(logger)
    if response == 1: return 1
    
    
    return 0


# experimental code, not used in production.


from optlang.glpk_interface import Model as Problem
import cobra



def test_manual_FBA(model, obid='biomass_LPL60'):
    
    
    # get the stoichiometric matrix (S): rows:metabolites, columns:reactions, cells:stoichiometric coefficients.
    # metabolite and reaction IDs are assumed unique (that is, no reapeated row ID or col ID).
    S = cobra.util.array.create_stoichiometric_matrix(model, array_type='DataFrame')
    
    
    # define the problem description in json format (dict)
    problem_json = {}  
    problem_json['name'] = 'manual_FBA'
    problem_json['variables'] = []
    problem_json['constraints'] = []
    problem_json['objective'] = {}
    
    
    # some convertions to speed up the table accession
    rid_to_index = {rid: index for index, rid in enumerate(S.columns)}
    mid_to_index = {mid: index for index, mid in enumerate(S.index)}
    S = S.values  # convert to numpy.ndarray  (faster cell accession)


    # define Variables, that are fluxes through reactions:
    # (variables are constrained with a LB and an UB)
    for rid in rid_to_index.keys():
        r = model.reactions.get_by_id(rid)
        problem_json['variables'].append({'name': rid, 'lb': r.lower_bound, 'ub': r.upper_bound, 'type': 'continuous'})
        
    
    # define Constraints, that are mass balances for each metabolite:
    # (in steady state, each mass balance equals to 0)
    for mid, index_r in mid_to_index.items():  # iterate metabolites
        add_args = []
        
        
        for rid, index_c in rid_to_index.items():  # iterate reactions
            stoich_coeff = S[index_r, index_c]   # get the stoichiometric coefficient from S.
            if stoich_coeff == 0: continue   # would be a useless term (save time).
            
            
            # define the addend (contribution of this reaction to the current mid):
            add_args.append({'type': 'Mul', 'args': [{'type': 'Number', 'value': stoich_coeff}, {'type': 'Symbol', 'name': rid}]})
        # add final mass balance for this mid:
        problem_json['constraints'].append({'name': f'{mid}_constraint', 'expression': {'type': 'Add', 'args': add_args}, 'lb': 0, 'ub': 0, 'indicator_variable': None, 'active_when': 1})
    
        
    # finally define the objective (1*obid): 
    obdir = 'max'  # minimize or maximize
    problem_json['objective'] = {'name': 'my_objective', 'expression': {'type': 'Mul', 'args': [{'type': 'Number', 'value': 1.0}, {'type': 'Symbol', 'name': obid}]}, 'direction': obdir}
    
    
    # convert json/dict to optlang problem:
    problem = Problem.from_json(problem_json)
    status = problem.optimize()
    return problem.objective.value



def test_prio_gapfilling(model, Rset_dataframe, biom_id='Growth', minflux=0.01, ):
    
    
    # get the stoichiometric matrix (S): rows:metabolites, columns:reactions, cells:stoichiometric coefficients.
    # metabolite and reaction IDs are assumed unique (that is, no reapeated row ID or col ID).
    S = cobra.util.array.create_stoichiometric_matrix(model, array_type='DataFrame')
    
    
    # define the problem description in json format (dict)
    problem_json = {}  
    problem_json['name'] = 'prio_gapfilling'
    problem_json['variables'] = []
    problem_json['constraints'] = []
    problem_json['objective'] = {}
    
    
    # some convertions to speed up the table accession
    rid_to_index = {rid: index for index, rid in enumerate(S.columns)}
    mid_to_index = {mid: index for index, mid in enumerate(S.index)}
    S = S.values  # convert to numpy.ndarray  (faster cell accession)
    
    
    # define the metabolic reactions that are not part of the draft panmodel:
    excluding = set(Rset_dataframe.index.to_list())


    # define Variables, that are fluxes through reactions:
    # (variables are constrained with a LB and an UB)
    for rid in rid_to_index.keys():
        r = model.reactions.get_by_id(rid)
        lb = r.lower_bound
        ub = r.upper_bound
        if rid == biom_id: lb = minflux  # apply minflux for biomass formation
        problem_json['variables'].append({'name': rid, 'lb': lb, 'ub': ub, 'type': 'continuous'})
        if rid in excluding:  # in addition to the flux variable, add another binar variable (indicator variable).
            problem_json['variables'].append({'name': f'binary_{rid}', 'lb': 0, 'ub': 1, 'type': 'binary'})
        

    # define Constraints, that are mass balances for each metabolite:
    # (in steady state, each mass balance equals to 0)
    for mid, index_r in mid_to_index.items():  # iterate metabolites
        add_args = []
        
        
        for rid, index_c in rid_to_index.items():  # iterate reactions
            stoich_coeff = S[index_r, index_c]   # get the stoichiometric coefficient from S.
            if stoich_coeff == 0: continue   # would be a useless term (save time).
            
            
            # define the addend (contribution of this reaction to the current mid):
            add_args.append({'type': 'Mul', 'args': [{'type': 'Number', 'value': stoich_coeff}, {'type': 'Symbol', 'name': rid}]})
        # add final mass balance for this mid:
        problem_json['constraints'].append({'name': f'{mid}_constraint', 'expression': {'type': 'Add', 'args': add_args}, 'lb': 0, 'ub': 0, 'indicator_variable': None, 'active_when': 1})
   
        
    # Define additional Constraints, using the indicator variables.
    # Contraints defined here: yi * LBi <= vi <= yi * UBi  (for each reaction i in 'excluding')
    # Decompose the two terms: 
    # yi * LBi <= vi   ===>   vi - yi * LBi >= 0    
    # vi <= yi * UBi   ===>   vi - yi * UBi <= 0
    # Where: --- yi is the indicator variable for the reaction i
    #        --- LBi, UBi are lower- and upper bounds for the reaction i
    #        --- vi is the flux variable through reaction i 
    #        --- i are potential gap-filler, ie reactions that are not inlcuded in the draft.
    for rid, row in Rset_dataframe.iterrows(): 
        
        
        # vi - yi * LBi >= 0    ===dissecting===>  [vi]  [- yi * LBi]  [>= 0]
        add_args = []
        add_args.append({'type': 'Mul', 'args': [{'type': 'Number', 'value': 1}, {'type': 'Symbol', 'name': rid}]})  #  vi
        add_args.append({'type': 'Mul', 'args': [{'type': 'Number', 'value': -row['lb']}, {'type': 'Symbol', 'name': f'binary_{rid}'}]})  # - yi * LBi
        problem_json['constraints'].append({'name': f'indicator_{rid}_lb', 'expression': {'type': 'Add', 'args': add_args}, 'lb': 0, 'ub': 1000, 'indicator_variable': None, 'active_when': 1})  # >= 0
        
        
        # vi - yi * UBi <= 0    ===dissecting===>  [vi]  [- yi * UBi]  [<= 0]
        add_args = []
        add_args.append({'type': 'Mul', 'args': [{'type': 'Number', 'value': 1}, {'type': 'Symbol', 'name': rid}]})  #  vi
        add_args.append({'type': 'Mul', 'args': [{'type': 'Number', 'value': -row['ub']}, {'type': 'Symbol', 'name': f'binary_{rid}'}]})  # - yi * UBi
        problem_json['constraints'].append({'name': f'indicator_{rid}_ub', 'expression': {'type': 'Add', 'args': add_args}, 'lb': -1000, 'ub': 0, 'indicator_variable': None, 'active_when': 1})  # <= 0
    
        
    # finally define the objective :  sum_i ( (1/(1+n_score))*yi )
    obdir = 'min'  # minimize or maximize
    add_args = []
    for rid, row in Rset_dataframe.iterrows(): 
        penalty = 1 / (1 + row['n_score'])  # inversely proportional to the normalized score
        add_args.append({'type': 'Mul', 'args': [{'type': 'Number', 'value': penalty}, {'type': 'Symbol', 'name': f'binary_{rid}'}]}) 
        
        
    problem_json['objective'] = {'name': 'my_objective', 'expression': {'type': 'Add', 'args': add_args}, 'direction': obdir}
    
    
    # convert json/dict to optlang problem:
    problem = Problem.from_json(problem_json)
    return problem

    """
    status = problem.optimize()
    return problem.objective.value
    """


import os
import shutil 
import subprocess
import pickle
import multiprocessing
import itertools


import pandas as pnd


from ..commons import chunkize_items
from ..commons import load_the_worker
from ..commons import gather_results
from ..commons import check_cached
from ..commons import create_summary
from ..commons import extract_aa_seq_from_genome
from ..commons import get_blast_header



def alignment_to_couples(accession, cluster_to_relfreq, seq_to_cluster): 
    
    
    # read the alignment with extra columns: 
    colnames = f'{get_blast_header()}'.split(' ')
    alignment = pnd.read_csv(f'working/rec_broken/alignments/{accession}.tsv', sep='\t', names=colnames )
    alignment['qcov'] = round((alignment['qend'] -  alignment['qstart'] +1)/ alignment['qlen'] * 100, 1)
    alignment['scov'] = round((alignment['send'] -  alignment['sstart'] +1)/ alignment['slen'] * 100, 1)
    
    
    # filter and sort by evalue (sorting is important because we consider each CDS just 1 time)
    alignment = alignment[alignment['pident'] >= 90]
    alignment = alignment[alignment['qcov'] >= 70]
    alignment = alignment[alignment['evalue'] <= 1e-5]
    alignment = alignment.sort_values('evalue', ascending=True)
    alignment = alignment.reset_index(drop=True)


    # create the 'prognum' (progressive number) column:
    alignment['prognum'] = None
    for index, row in alignment.iterrows(): 
        alignment.loc[index, 'prognum'] = int(row['qseqid'].split('_', 1)[1])


    # create a blacklist because each CDS must be taken just 1 time: 
    blacklist = set()


    # group hsps by cluster:
    df_couples = []
    groups = alignment.groupby('sseqid').groups
    for cluster in groups.keys():
        hsps = alignment.iloc[ groups[cluster], ]
        # if a CDS is repeated for the same cluster, take the best hit.
        hsps = hsps.drop_duplicates(subset='qseqid', keep='first')
        # get the length of the representative seq (it's constant)
        slen = hsps['slen'].values[0] 


        # search for couples having progressive numbers:
        prognums = list(hsps['prognum'].values)
        for i in prognums:
            if i in blacklist: continue
            if i+1 in prognums and i+2 not in prognums and i-1 not in prognums:
                # get the two pieces of this protein: 
                good_couple = hsps[hsps['prognum'].isin([i, i+1])]   
                # below we compute several metrics: 

                
                # overall coverage % (include the gap between the two pieces):
                overall_cov = ( max(good_couple['send']) - min(good_couple['sstart']) +1 ) / slen * 100

                # superimposition % between the two pieces: 
                if min(good_couple['send']) >= max(good_couple['sstart']):
                    sup = ( min(good_couple['send']) - max(good_couple['sstart']) +1 ) / slen * 100
                else: sup = 0

                # compute the gap % between the two pieces:
                if min(good_couple['send']) < max(good_couple['sstart']):
                    gap = ( max(good_couple['sstart']) - min(good_couple['send']) -1 ) / slen * 100
                else: gap = 0

                # compute the query len relative to the subject (constant)
                rqlen1 = good_couple['qlen'].values[0] / slen * 100
                rqlen2 = good_couple['qlen'].values[1] / slen * 100

                # get the relative frequencies:
                relfreq_cluster = cluster_to_relfreq[cluster]
                relfreq_piece1 = cluster_to_relfreq[seq_to_cluster[good_couple['qseqid'].values[0]]]
                relfreq_piece2 = cluster_to_relfreq[seq_to_cluster[good_couple['qseqid'].values[1]]]

                
                # if this couple respect the thresholds: 
                if  overall_cov >= 70 and \
                    sup <= 30 and \
                    rqlen1 < 90 and rqlen2 < 90 and \
                    relfreq_cluster > relfreq_piece1 and relfreq_cluster > relfreq_piece2:
                    
                    
                    # include in results and update the blacklist:
                    df_couples.append(good_couple)
                    blacklist.add(i)
                    blacklist.add(i+1)

                    
                # consider just the first progressive couple (that is, with the highest evalues):
                break  


    # write results to disk: 
    if df_couples == []:  # create an empty dataframe
        df_couples = [pnd.DataFrame(columns=alignment.columns)]
    df_couples = pnd.concat(df_couples, axis=0)
    df_couples = df_couples.drop('prognum', axis=1)
    df_couples = df_couples.reset_index(drop=True)
    df_couples.to_csv(f'working/rec_broken/couples/{accession}.csv')
    
    
    return df_couples


    
def get_updated_column(pam, accession, df_couples, cluster_to_relfreq, seq_to_cluster):
    
    
    # this module will release a new updated pam.
    # the new pam will be created gluing together singular columns.
    # columns will be first translated (.T) to be compliant with the .commons lib.
    pam_column = pam.loc[: , [accession]].copy()
    
    
    # save all the new mappings
    edits_dict = {}
    
    
    # parse each couple to trace the jumping of protein pieces:
    with open(f'working/rec_broken/edits/{accession}.txt', "w") as w_handler:
        groups = df_couples.groupby('sseqid').groups
        for cluster in groups.keys():
            couple = df_couples.iloc[ groups[cluster], ]


            # convert the original cell of the gained cluster to set
            ori_cell = pam_column.loc[cluster, accession]
            if type(ori_cell)==float: ori_cell_set = set()  # empty cell
            else: ori_cell_set = set(ori_cell.split(';'))


            # get the two pieces (prefix and progressive number)
            cds_ids = couple['qseqid'].to_list()
            cds_prognums = sorted([i.rsplit('_', 1)[1] for i in cds_ids]) 
            prefix = cds_ids[0].split('_', 1)[0]


            # remove eventual pieces from the gained cluster cell, then add the glued protein
            ori_cell_depleted = ori_cell_set - set(cds_ids)
            new_cell = set([f'{prefix}_frag_' + '_'.join(cds_prognums)]).union(ori_cell_depleted)
            new_cell = ';'.join(new_cell)

            
            # get the freq of the gained cluster, log the change, and apply to pam
            rel_freq = cluster_to_relfreq[cluster]
            print(f'{cluster} ({rel_freq}%): {ori_cell} --> {new_cell}', file=w_handler)
            pam_column.loc[cluster, accession] = new_cell


            # now update the cells of the two pieces, following the same logic above.
            for cds in cds_ids: 
                cluster2 = seq_to_cluster[cds]
                ori_cell2 = pam_column.loc[cluster2, accession]
                ori_cell2_set = set() if type(ori_cell2)==float else set(ori_cell2.split(';'))
                new_cell2 = set(ori_cell2_set) - set(cds_ids) 
                new_cell2 = ';'.join(new_cell2)
                rel_freq2 = cluster_to_relfreq[cluster2]
                print(f'\t{cluster2} ({rel_freq2}%): {ori_cell2} --> {new_cell2}', file=w_handler)
                pam_column.loc[cluster2, accession] = new_cell2
                
                
            # new: create a new seq_to_cluster dictionary
            for cds in cds_ids: 
                edits_dict[f'{cds}:{seq_to_cluster[cds]}'] = f'{prefix}_frag_' + '_'.join(cds_prognums) + f':{cluster}'
    
    
    # save the dictionary: 
    with open(f'working/rec_broken/edits/{accession}.pickle', 'wb') as file:
        pickle.dump(edits_dict, file)
    
    
    return pam_column[accession]



def task_recbroken(genome, args):
    
    
    # retrive the arguments:
    pam = args['pam']
    cluster_to_relfreq = args['cluster_to_relfreq']
    seq_to_cluster = args['seq_to_cluster']
    
    
    # get the accession and proteome file:
    basename = os.path.basename(genome)
    accession, _ = os.path.splitext(basename)
    proteome = f'working/proteomes/{accession}.faa'
    
    
    # create a database for later extraction of recovered sequences: 
    os.makedirs(f'working/rec_broken/databases/{accession}/', exist_ok=True)
    shutil.copyfile(genome, f'working/rec_broken/databases/{accession}/{accession}.fna')  # just the content, not the permissions.
    command = f"""makeblastdb -in working/rec_broken/databases/{accession}/{accession}.fna -dbtype nucl -parse_seqids""" # '-parse_seqids' is required for 'blastdbcmd'.
    process = subprocess.Popen(command, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    process.wait()
    
    
    # perform the blastp: proteins on representatives:  
    command = f'''blastp \
        -query {proteome} \
        -db working/rec_broken/representatives/representatives.ren.faa \
        -out working/rec_broken/alignments/{accession}.tsv \
        -outfmt "6 {get_blast_header()}"
    '''
    process = subprocess.Popen(command, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    process.wait()


    # parse the alignment to get the good couples.
    df_couples = alignment_to_couples(accession, cluster_to_relfreq, seq_to_cluster)
    
    
    # parse the good couples to update the pam column:
    pam_column = get_updated_column(pam, accession, df_couples, cluster_to_relfreq, seq_to_cluster)

    
    # return new rows for load_the_worker():
    row = pam_column.to_dict()
    row['accession'] = accession
    return [row]
    
    

def populate_results_df(logger):
    
    
    # load the previously created dictionaries: 
    with open('working/proteomes/species_to_proteome.pickle', 'rb') as handler:
        species_to_proteome = pickle.load(handler)
    with open('working/coordinates/seq_to_coords.pickle', 'rb') as handler:
        seq_to_coords = pickle.load(handler)
        
    
    # parse each accession:
    for species in species_to_proteome.keys(): 
        for proteome in species_to_proteome[species]:
            basename = os.path.basename(proteome)
            accession, _ = os.path.splitext(basename)
            results_df = []
            
            
            # get the couple
            df_couples = pnd.read_csv(f'working/rec_broken/couples/{accession}.csv', index_col=0)
            groups = df_couples.groupby('sseqid').groups
            for cluster in groups.keys():
                couple = df_couples.iloc[ groups[cluster], ]
                
                
                # get the new sequence id
                cds_ids = couple['qseqid'].to_list()
                cds_prognums = sorted([i.rsplit('_', 1)[1] for i in cds_ids]) 
                prefix = cds_ids[0].split('_', 1)[0]
                new_seq_id = f'{prefix}_frag_' + '_'.join(cds_prognums)
                
                
                # get the accession
                couple_accessions = set([seq_to_coords[seq]['accession'] for seq in cds_ids])
                if len(couple_accessions) != 1:
                    logger.error(f"Found different accessions in this couple: {cds_ids}.")
                    return 1
                couple_accession = list(couple_accessions)[0]
                
                
                # get the contig
                couple_contigs = set([seq_to_coords[seq]['contig'] for seq in cds_ids])
                if len(couple_contigs) != 1:
                    logger.info(f"Found different contigs in this couple: {cds_ids} (accession {couple_accession}).")
                    #return 1
                couple_contig = list(couple_contigs)[0]
                
                
                # get the strand
                couple_strands = set([seq_to_coords[seq]['strand'] for seq in cds_ids])
                if len(couple_strands) != 1:
                    logger.info(f"Found different strands in this couple: {cds_ids} (accession {couple_accession}).")
                    #return 1
                couple_strand = list(couple_strands)[0]
                
                
                # get the start - end
                start_1 = seq_to_coords[cds_ids[0]]['start']
                start_2 = seq_to_coords[cds_ids[1]]['start']
                end_1 = seq_to_coords[cds_ids[0]]['end']
                end_2 = seq_to_coords[cds_ids[1]]['end']
                if start_1 > end_1: 
                    logger.info(f"Found start_1 < end_1 in this couple: {cds_ids} (accession {couple_accession}).")
                    #return 1
                if start_2 > end_2: 
                    logger.info(f"Found start_2 < end_2 in this couple: {cds_ids} (accession {couple_accession}).")
                    #return 1
                couple_start = min([start_1, start_2])
                couple_end = max([end_1, end_2])
                
                
                # write the results dataframe
                results_df.append({
                    'ID': new_seq_id, 'cluster': cluster, 
                    'accession': couple_accession, 'contig': couple_contig, 'strand': couple_strand, 
                    'start': couple_start, 'end': couple_end
                })
            results_df = pnd.DataFrame.from_records(results_df)
            results_df.to_csv(f'working/rec_broken/results/{accession}.csv')
    
    
    return 0
    
    

def update_seq_to_coords(logger): 
    
    
    # load the previously created species_to_proteome: 
    with open('working/proteomes/species_to_proteome.pickle', 'rb') as handler:
        species_to_proteome = pickle.load(handler)
        
    
    # get the good accessions (passing the quality filters):
    good_accessions = []
    for species in species_to_proteome.keys(): 
        for proteome in species_to_proteome[species]:
            basename = os.path.basename(proteome)
            accession, _ = os.path.splitext(basename)
            good_accessions.append(accession)
    
    
    # parse the couples/log files to get the seqs ID to erease:
    to_erease = []
    for accession in good_accessions: 
        df_couples = pnd.read_csv(f'working/rec_broken/couples/{accession}.csv', index_col=0)
        to_erease = to_erease + df_couples['qseqid'].to_list()
    
        
    # create an updateed seq_to_coords dict: 
    with open('working/coordinates/seq_to_coords.pickle', 'rb') as handler:
        seq_to_coords = pickle.load(handler)
    logger.debug(f'rec_broken: seq_to_coords: starting from {len(seq_to_coords.values())} sequences.')
    seq_to_coords_update = {}
        
    
    # remove seqs from the updated dictionary:
    for seq in seq_to_coords.keys(): 
        attribs = seq_to_coords[seq] 
        if attribs['accession'] not in good_accessions: 
            continue  # remove all seq IDs belonging to accessions not passing the quality filter. 
        if seq in to_erease: 
            continue  # remove seq IDs belonging to the good couples discovered by this module. 
        seq_to_coords_update[seq] = attribs
    logger.debug(f'rec_broken: seq_to_coords: {len(seq_to_coords_update.values())} after removing filtered accessions and frag couples.')

        
    # now add the new seqs (recovered by this module): 
    for accession in good_accessions: 
        results_df = pnd.read_csv(f'working/rec_broken/results/{accession}.csv', index_col=0)
        for index, row in results_df.iterrows():
            seq_to_coords_update[row['ID']] = {'accession': row['accession'], 'contig': row['contig'], 'strand': row['strand'], 'start': row['start'], 'end': row['end']}
    logger.debug(f'rec_broken: seq_to_coords: {len(seq_to_coords_update.values())} sequences after the addition of new IDs.')

    
    # save the update dictionary: 
    with open('working/rec_broken/seq_to_coords.pickle', 'wb') as file:
        pickle.dump(seq_to_coords_update, file)
        
        
        
def join_edits_dict(logger):
    
    
    # load the previously created species_to_proteome: 
    with open('working/proteomes/species_to_proteome.pickle', 'rb') as handler:
        species_to_proteome = pickle.load(handler)
        
    
    # get the good accessions (passing the quality filters):
    good_accessions = []
    for species in species_to_proteome.keys(): 
        for proteome in species_to_proteome[species]:
            basename = os.path.basename(proteome)
            accession, _ = os.path.splitext(basename)
            good_accessions.append(accession)
    
    
    # parse the couples/log files to get the seqs ID to erease:
    edits_dict = {}
    for accession in good_accessions: 
        with open(f'working/rec_broken/edits/{accession}.pickle', 'rb') as handler:
            curr_edits = pickle.load(handler)
        for key, value in curr_edits.items():
            edits_dict[key] = value
            
            
    # save the dictionary: 
    with open(f'working/rec_broken/edits_dict.pickle', 'wb') as file:
        pickle.dump(edits_dict, file)
        
        

def update_sequences(logger): 
    
    
    # load the previously created species_to_proteome: 
    with open('working/proteomes/species_to_proteome.pickle', 'rb') as handler:
        species_to_proteome = pickle.load(handler)
        
    
    # get the accessions to work with: 
    accessions = []
    for species in species_to_proteome.keys(): 
        for proteome in species_to_proteome[species]:
            basename = os.path.basename(proteome)
            accession, _ = os.path.splitext(basename)
            accessions.append(accession)
            
    
    # parse the couples/log files to get the seqs ID to erease:
    to_erease = []
    for accession in accessions: 
        df_couples = pnd.read_csv(f'working/rec_broken/couples/{accession}.csv', index_col=0)
        to_erease = to_erease + df_couples['qseqid'].to_list()
            
            
    # create an updated df, removing the seqs to erease:
    sequences_df = pnd.read_csv('working/clustering/sequences.csv' , index_col=0)
    logger.debug(f'rec_broken: sequences dataframe: starting from {len(sequences_df)} sequences.')
    sequences_df_updated = sequences_df.copy()
    sequences_df_updated = sequences_df_updated.drop(to_erease)
    logger.debug(f'rec_broken: sequences dataframe: {len(sequences_df_updated)} after removing frag couples.')
    
    
    # now add the new seqs
    new_rows = []
    for accession in accessions: 
        results_df = pnd.read_csv(f'working/rec_broken/results/{accession}.csv', index_col=0)
        for index, row in results_df.iterrows():
            contig = row['contig']
            strand = row['strand']
            start = row['start']
            end = row['end']
            seq, seq_tostop = extract_aa_seq_from_genome(
                f'working/rec_broken/databases/{accession}/{accession}.fna', 
                contig, strand, start, end)
            new_rows.append({'cds': row['ID'], 'accession': accession, 'aaseq': seq})
    new_rows = pnd.DataFrame.from_records(new_rows)
    new_rows = new_rows.set_index('cds', drop=True, verify_integrity=True)
    sequences_df_updated = pnd.concat([sequences_df_updated, new_rows])
    logger.debug(f'rec_broken: seq_to_coords: {len(sequences_df_updated)} sequences after the addition of new IDs.')

    
    # save the update version:
    sequences_df_updated.to_csv('working/rec_broken/sequences.csv')
    


def recovery_broken(logger, cores):
    
    
    # some log messages:
    logger.info("Recovering the proteins broken in two pieces...")
    
    
    # create sub-directories without overwriting:
    os.makedirs('working/rec_broken/', exist_ok=True)
    os.makedirs('working/rec_broken/edits/', exist_ok=True)
    os.makedirs('working/rec_broken/representatives/', exist_ok=True)
    os.makedirs('working/rec_broken/databases/', exist_ok=True)
    os.makedirs(f'working/rec_broken/alignments/', exist_ok=True)
    os.makedirs('working/rec_broken/couples/', exist_ok=True)
    os.makedirs('working/rec_broken/results/', exist_ok=True)
    
    
    # check if it's everything pre-computed
    response = check_cached(
        logger, pam_path='working/rec_broken/pam.csv',
        summary_path='working/rec_broken/summary.csv',
        imp_files = [
            'working/rec_broken/sequences.csv',
            'working/rec_broken/seq_to_coords.pickle',
            'working/rec_broken/edits_dict.pickle',])
    if response == 0: 
        return 0
    

    # copy representative sequences (all) and make a database
    shutil.copyfile(f'working/clustering/representatives.ren.faa', f'working/rec_broken/representatives/representatives.ren.faa') 
    command = f"""makeblastdb -in working/rec_broken/representatives/representatives.ren.faa -dbtype prot"""
    process = subprocess.Popen(command, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    process.wait()
    
    
    # load the assets to form the args dictionary:
    pam = pnd.read_csv('working/clustering/pam.csv', index_col=0)
    with open('working/clustering/cluster_to_relfreq.pickle', 'rb') as handler:
        cluster_to_relfreq = pickle.load(handler)
    with open('working/clustering/seq_to_cluster.pickle', 'rb') as handler:
        seq_to_cluster = pickle.load(handler)
    
    
    # load the previously created species_to_proteome: 
    with open('working/genomes/species_to_genome.pickle', 'rb') as handler:
        species_to_genome = pickle.load(handler)
        
        
    # create items for parallelization: 
    items = []
    for species in species_to_genome.keys(): 
        for genome in species_to_genome[species]: 
            items.append(genome)
            
            
    # randomize and divide in chunks: 
    chunks = chunkize_items(items, cores)
    
    
    # initialize the globalpool:
    globalpool = multiprocessing.Pool(processes=cores, maxtasksperchild=1)
    
    
    # start the multiprocessing: 
    results = globalpool.imap(
        load_the_worker, 
        zip(chunks, 
            range(cores), 
            itertools.repeat(['accession'] + list(pam.T.columns)), 
            itertools.repeat('accession'), 
            itertools.repeat(logger), 
            itertools.repeat(task_recbroken),  # will return a new updated pam.
            itertools.repeat({'pam': pam, 'cluster_to_relfreq': cluster_to_relfreq, 'seq_to_cluster': seq_to_cluster, }),
        ), chunksize = 1)
    all_df_combined = gather_results(results)
    
    
    # empty the globalpool
    globalpool.close() # prevent the addition of new tasks.
    globalpool.join() 
    
    
    # get the updated pam, and remove empty rows:
    pam_updated = all_df_combined.T
    pam_updated = pam_updated.replace({'': None})
    empty_rows_bool = pam_updated.isnull().all(axis=1)
    empty_rows_df = pam_updated.loc[empty_rows_bool]
    logger.debug(f"rec_broken: ended up with {len(empty_rows_df)} devoided clusters.")
    pam_updated = pam_updated.drop(empty_rows_df.index)
    pam_updated.to_csv('working/rec_broken/pam.csv')
    
    
    # get the results dataframe following conventions
    response = populate_results_df(logger)
    if response == 1: return 1

    
    # update the squence to coordinates dict (removing filtered genomes, and protein frags)
    update_seq_to_coords(logger)
    
    
    # update the sequences dataframe (removing protein frags)
    update_sequences(logger)
    
    
    # join together all the strain specific edits dicts
    join_edits_dict(logger)
    
    
    # create a summary for this module reading the results dataframes: 
    create_summary(logger, module_dir='working/rec_broken/')
    

    
    return 0